"""
Model objects for mimic images.
"""

from characteristic import attributes
import uuid
import random


@attributes(['image_id', 'tenant_id', 'name', 'minDisk', 'minRam', 'image_size'])
class Image(object):
    """
    A Image object
    """

    is_default = False

    def set_is_default(self):
        """
        Sets the image as default
        """
        self.is_default = True

    @classmethod
    def image_id(cls):
        """
        Create a unique id for an image
        """
        return str(uuid.uuid4())

    @classmethod
    def image_size(cls):
        """
        Generate a random size for images
        """
        return random.randint(250000, 80000000000)

    def links_json(self, absolutize_url):
        """
        Create a JSON-serializable data structure describing the links to this
        image.
        """
        return [
            {
                "href": absolutize_url("v2/{0}/images/{1}"
                                       .format(self.tenant_id, self.image_id)),
                "rel": "self"
            },
            {
                "href": absolutize_url("{0}/images/{1}"
                                       .format(self.tenant_id, self.image_id)),
                "rel": "bookmark"
            },
            {
                "href": absolutize_url("/images/{0}"
                                       .format(self.image_id)),
                "type": "application/vnd.openstack.image",
                "rel": "alternate"
            }
        ]

    def brief_json(self, absolutize_url):
        """
        Brief JSON-serializable version of this flavor, for the non-details
        list flavors request.
        """
        template = {}
        template.update({
            "id": self.image_id,
            "links": self.links_json(absolutize_url),
            "name": self.name
        })
        return template

    def detailed_json(self, absolutize_url):
        """
        Long-form JSON-serializable object representation of this flavor, as
        returned by either a GET on this individual flavor or a member in the
        list returned by the list-details request.
        """
        template = {}
        template.update({
            "id": self.image_id,
            "links": self.links_json(absolutize_url),
            "name": self.name,
            "minRam": self.minRam,
            "minDisk": self.minDisk,
            "OS-EXT-IMG-SIZE:size": self.image_size,
            "com.rackspace__1__ui_default_show": self.is_default,
            "created": "1972-01-01_15-59-11",
            "updated": "1972-01-01_15-59-11",
            "progress": 100,
            "status": "ACTIVE",
            "metadata": self.metadata_json()
        })
        return template


class RackspaceWindowsImage(Image):
    """
    A Rackspace window image object representation
    """
    images = {"Windows Server 2008 R2 SP1 + SQL Server 2008 R2 SP2 Standard":
              {"minRam": 2048, "minDisk": 40, "OS-EXT-IMG-SIZE:size": Image.image_size(),
               "id": Image.image_id(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2008 R2 SP1 + SQL Server 2008 R2 SP2 Web":
              {"minRam": 2048, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2008 R2 SP1":
              {"minRam": 1024, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2012 R2":
              {"minRam": 1024, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2012 + SQL Server 2012 SP1 Web":
              {"minRam": 2048, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2012 + SQL Server 2012 SP1 Standard":
              {"minRam": 2048, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2012 R2 + SQL Server 2014 Standard":
              {"minRam": 2048, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2012 R2 + SQL Server 2014 Web":
              {"minRam": 2048, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"},
              "Windows Server 2012":
              {"minRam": 1024, "minDisk": 40, "id": Image.image_id(),
               "OS-EXT-IMG-SIZE:size": Image.image_size(), "com.rackspace__1__ui_default_show": "True"}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "windows",
            "vm_mode": "",
            "auto_disk_config": "disabled",
            "org.openstack__1__os_distro": "com.microsoft.server"
        }


class RackspaceArchImage(Image):
    """
    A Rackspace Arch image object representation
    """
    images = {"Arch 2015.7 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                      "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                      "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.archlinux",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceCentOSPVHMImage(Image):
    """
    A Rackspace CentOS HVM image object representation
    """
    images = {"CentOS 7 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                   "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                   "com.rackspace__1__ui_default_show": "True", "id": Image.image_id()},
              "CentOS 6 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                   "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                   "com.rackspace__1__ui_default_show": "True", "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.centos",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceCentOSPVImage(Image):
    """
    A Rackspace CentOS Xen image object representation
    """
    images = {"CentOS 6 (PV)": {"minRam": 512, "minDisk": 20, "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                "id": Image.image_id()},
              "CentOS 5 (PV)": {"minRam": 512, "minDisk": 20, "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!io1,!memory1,!compute1,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.centos",
            "vm_mode": "xen",
            "auto_disk_config": "True"
        }


class RackspaceCoreOSImage(Image):
    """
    A Rackspace CoreOS image object representation
    """
    images = {"CoreOS (Beta)": {"minRam": 512, "minDisk": 20, "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                "id": Image.image_id()},
              "CoreOS (Alpha)": {"minRam": 512, "minDisk": 20,
                                 "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                 "id": Image.image_id()},
              "CoreOS (Stable)": {"minRam": 512, "minDisk": 20,
                                  "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                  "com.rackspace__1__ui_default_show": "True", "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.coreos",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceDebianImage(Image):
    """
    A Rackspace Debian image object representation
    """
    images = {"Debian 7 (Wheezy) (PVHVM)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                            "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                            "com.rackspace__1__ui_default_show": "True"},
              "Debian Unstable (Sid) (PVHVM)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                                "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "Debian Testing (Stretch) (PVHVM)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                                   "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "Debian 8 (Jessie) (PVHVM)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                            "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                            "com.rackspace__1__ui_default_show": "True"}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.debian",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceFedoraImage(Image):
    """
    A Rackspace Fedora image object representation
    """
    images = {"Fedora 21 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                    "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                    "com.rackspace__1__ui_default_show": "True",
                                    "id": Image.image_id()},
              "Fedora 22 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                    "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                    "com.rackspace__1__ui_default_show": "True",
                                    "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.fedoraproject",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceFreeBSDImage(Image):
    """
    A Rackspace FreeBSD image object representation
    """
    images = {"FreeBSD 10 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                     "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                     "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.freebsd",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceGentooImage(Image):
    """
    A Rackspace Gentoo image object representation
    """
    images = {"Gentoo 15.3 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                      "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                      "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.gentoo",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceOpenSUSEImage(Image):
    """
    A Rackspace OpenSUSE image object representation
    """
    images = {"OpenSUSE 13.2 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                        "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                        "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.opensuse",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceRedHatPVHMImage(Image):
    """
    A Rackspace Red Hat image object representation
    """
    images = {"Red Hat Enterprise Linux 7 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                                     "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                                     "com.rackspace__1__ui_default_show": "True",
                                                     "id": Image.image_id()},
              "Red Hat Enterprise Linux 6 (PVHVM)": {"minRam": 512, "minDisk": 20,
                                                     "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                                     "com.rackspace__1__ui_default_show": "True",
                                                     "id": Image.image_id()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.redhat",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceRedHatPVImage(Image):
    """
    A Rackspace Red Hat image object representation
    """
    images = {"Red Hat Enterprise Linux 6 (PV)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                                  "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "Red Hat Enterprise Linux 5 (PV)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                                  "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!io1,!memory1,!compute1,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.redhat",
            "vm_mode": "xen",
            "auto_disk_config": "True"
        }


class RackspaceScientificImage(Image):
    """
    A Rackspace Scientific image object representation
    """
    images = {"Scientific Linux 6 (PVHVM)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                             "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "Scientific Linux 7 (PVHVM)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                             "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.scientificlinux",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceUbuntuPVHMImage(Image):
    """
    A Rackspace Ubuntu image object representation
    """
    images = {"Ubuntu 15.04 (Vivid Vervet) (PVHVM)": {"minRam": 512, "minDisk": 20,
                                                      "id": Image.image_id(),
                                                      "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                                      "com.rackspace__1__ui_default_show": "True"},
              "Ubuntu 12.04 LTS (Precise Pangolin) (PVHVM)":
                  {"minRam": 512, "minDisk": 20, "OS-EXT-IMG-SIZE:size": Image.image_size(),
                   "id": Image.image_id(), "com.rackspace__1__ui_default_show": "True"},
              "Ubuntu 14.04 LTS (Trusty Tahr) (PVHVM)": {"minRam": 512, "minDisk": 20,
                                                         "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                                         "id": Image.image_id(),
                                                         "com.rackspace__1__ui_default_show": "True"}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.ubuntu",
            "vm_mode": "hvm",
            "auto_disk_config": "disabled"
        }


class RackspaceUbuntuPVImage(Image):
    """
    A Rackspace Ubuntu image object representation
    """
    images = {"Ubuntu 12.04 LTS (Precise Pangolin) (PV)": {"minRam": 512, "minDisk": 20,
                                                           "id": Image.image_id(),
                                                           "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "Ubuntu 14.04 LTS (Trusty Tahr) (PV)": {"minRam": 512, "minDisk": 20,
                                                      "id": Image.image_id(),
                                                      "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!io1,!memory1,!compute1,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.ubuntu",
            "vm_mode": "xen",
            "auto_disk_config": "True"
        }


class RackspaceVyattaImage(Image):
    """
    A Rackspace Vyatta image object representation
    """
    images = {"Vyatta Network OS 6.7R9": {"minRam": 1024, "minDisk": 20, "id": Image.image_id(),
                                          "OS-EXT-IMG-SIZE:size": Image.image_size(),
                                          "com.rackspace__1__ui_default_show": "True"}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "*,!io1,!memory1,!compute1,!onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.vyatta",
            "vm_mode": "xen",
            "auto_disk_config": "False"
        }


@attributes(['image_id', 'tenant_id', 'name', 'minDisk', 'minRam', 'image_size'])
class OnMetalImage(object):
    """
    A Image object
    """

    is_default = False

    def set_is_default(self):
        """
        Sets image as a default
        """
        self.is_default = True

    def links_json(self, absolutize_url):
        """
        Create a JSON-serializable data structure describing the links to this
        image.
        """
        return [
            {
                "href": absolutize_url("v2/{0}/images/{1}"
                                       .format(self.tenant_id, self.image_id)),
                "rel": "self"
            },
            {
                "href": absolutize_url("{0}/images/{1}"
                                       .format(self.tenant_id, self.image_id)),
                "rel": "bookmark"
            },
            {
                "href": absolutize_url("/images/{0}"
                                       .format(self.image_id)),
                "type": "application/vnd.openstack.image",
                "rel": "alternate"
            }
        ]

    def brief_json(self, absolutize_url):
        """
        Brief JSON-serializable version of this flavor, for the non-details
        list flavors request.
        """
        return {
            "id": self.image_id,
            "links": self.links_json(absolutize_url),
            "name": self.name
        }

    def detailed_json(self, absolutize_url):
        """
        Long-form JSON-serializable object representation of this flavor, as
        returned by either a GET on this individual flavor or a member in the
        list returned by the list-details request.
        """
        template = {}
        template.update({
            "id": self.image_id,
            "links": self.links_json(absolutize_url),
            "name": self.name,
            "minRam": self.minRam,
            "minDisk": self.minDisk,
            "OS-EXT-IMG-SIZE:size": self.image_size,
            "com.rackspace__1__ui_default_show": self.is_default,
            "created": "1972-01-01_15-59-11",
            "updated": "1972-01-01_15-59-11",
            "status": "ACTIVE",
            "progress": 100,
            "metadata": self.metadata_json()
        })
        return template


class RackspaceOnMetalCoreOSImage(OnMetalImage):
    """
    A Rackspace OnMetal image object representation
    """
    images = {"OnMetal - CoreOS (Alpha)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                           "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - CoreOS (Beta)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                          "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - CoreOS (Stable)": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                            "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.coreos",
            "vm_mode": "metal",
            "auto_disk_config": "disabled"
        }


class RackspaceOnMetalFedoraImage(OnMetalImage):
    """
    A Rackspace OnMetal image object representation
    """
    images = {"OnMetal - Fedora 22": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                      "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - Fedora 21": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                      "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.fedoraproject",
            "vm_mode": "metal",
            "auto_disk_config": "disabled"
        }


class RackspaceOnMetalDebianImage(OnMetalImage):
    """
    A Rackspace OnMetal image object representation
    """
    images = {"OnMetal - Debian Testing (Stretch)": {"minRam": 512, "minDisk": 20,
                                                     "id": Image.image_id(),
                                                     "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - Debian Unstable (Sid)": {"minRam": 512, "minDisk": 20,
                                                  "id": Image.image_id(),
                                                  "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - Debian 8 (Jessie)": {"minRam": 512, "minDisk": 20,
                                              "id": Image.image_id(),
                                              "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - Debian 7 (Wheezy)": {"minRam": 512, "minDisk": 20,
                                              "id": Image.image_id(),
                                              "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.debian",
            "vm_mode": "metal",
            "auto_disk_config": "disabled"
        }


class RackspaceOnMetalUbuntuImage(OnMetalImage):
    """
    A Rackspace OnMetal image object representation
    """
    images = {"OnMetal - Ubuntu 15.04 (Vivid Vervet)": {"minRam": 512, "minDisk": 20,
                                                        "id": Image.image_id(),
                                                        "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - Ubuntu 14.04 LTS (Trusty Tahr)": {"minRam": 512, "minDisk": 20,
                                                           "id": Image.image_id(),
                                                           "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - Ubuntu 12.04 LTS (Precise Pangolin)":
                  {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                   "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.ubuntu",
            "vm_mode": "metal",
            "auto_disk_config": "disabled"
        }


class RackspaceOnMetalCentOSImage(OnMetalImage):
    """
    A Rackspace OnMetal image object representation
    """
    images = {"OnMetal - CentOS 6": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                     "OS-EXT-IMG-SIZE:size": Image.image_size()},
              "OnMetal - CentOS 7": {"minRam": 512, "minDisk": 20, "id": Image.image_id(),
                                     "OS-EXT-IMG-SIZE:size": Image.image_size()}}

    def metadata_json(self):
        """
        Create a JSON-serializable data structure describing
        ``metadata`` for an image.
        """
        return {
            "flavor_classes": "onmetal",
            "image_type": "base",
            "os_type": "linux",
            "org.openstack__1__os_distro": "org.centos",
            "vm_mode": "metal",
            "auto_disk_config": "disabled"
        }


class ImageStore(object):
    """
    A store for images to share between nova_api and glance_api
    """
    _images_store = []

    @classmethod
    def create_image_store(cls, tenant_id):
        """
        Generates the data for each image in each image class
        """
        image_classes = [RackspaceWindowsImage, RackspaceArchImage, RackspaceCentOSPVImage,
                         RackspaceCentOSPVHMImage, RackspaceCoreOSImage, RackspaceDebianImage,
                         RackspaceFedoraImage, RackspaceFreeBSDImage, RackspaceGentooImage,
                         RackspaceOpenSUSEImage, RackspaceRedHatPVImage, RackspaceRedHatPVHMImage,
                         RackspaceUbuntuPVImage, RackspaceUbuntuPVHMImage, RackspaceVyattaImage,
                         RackspaceScientificImage, RackspaceOnMetalCentOSImage,
                         RackspaceOnMetalCoreOSImage, RackspaceOnMetalDebianImage,
                         RackspaceOnMetalFedoraImage, RackspaceOnMetalUbuntuImage]
        if len(cls._images_store) < 1:
            for image_class in image_classes:
                for image, image_spec in image_class.images.iteritems():
                    image_name = image
                    image_id = image_spec['id']
                    minRam = image_spec['minRam']
                    minDisk = image_spec['minDisk']
                    image_size = image_spec['OS-EXT-IMG-SIZE:size']
                    image = image_class(image_id=image_id, tenant_id=tenant_id,
                                        image_size=image_size, name=image_name, minRam=minRam,
                                        minDisk=minDisk)
                    if 'com.rackspace__1__ui_default_show' in image_spec:
                        image.set_is_default()
                    cls._images_store.append(image)
        return cls._images_store



"""
Plugin for Rackspace cloud load balancer mock.
"""
from mimic.rest.loadbalancer_api import LoadBalancerApi, LoadBalancerControlApi

loadbalancer = LoadBalancerApi()
loadbalancer_control = LoadBalancerControlApi(lb_api=loadbalancer)



"""
Canned response for Queue
"""


def queues_example(queue_name):
    """
    Create queue response example
    """
    queue_example = {"href": "/v1/queues/" + queue_name,
                     "name": queue_name}
    return queue_example


def add_queue(queue_id, queue_name, tenant_id, q_cache):
    """
    Returns response of a newly created queue with
    response code 201, and adds the new queue to the queue_cache.
    """
    q_cache[queue_id] = queues_example(queue_name)
    q_cache[queue_id].update({"tenant_id": tenant_id})
    return None, 201


def list_queues(tenant_id, q_cache):
    """
    Returns a list of queues with response code 200
    """
    return {'queues': q_cache.values() or []}, 200


def delete_queue(queue_name, q_cache):
    """
    Deletes a queue with response code 204
    """
    for key, val in q_cache.items():
        if queue_name == val['name']:
            del q_cache[key]
            return None, 204



"""
Interfaces for Mimic.
"""

from zope.interface import Attribute, Interface


class IAPIMock(Interface):
    """
    An :obj:`IAPIMock` provides an API.
    """

    def catalog_entries(tenant_id):  # pragma:nocover
        """
        Generate some :obj:`mimic.catalog.Entry` objects given the tenant ID.

        :param unicode tenant_id: the semi-internal tenant ID generated by
            Mimic.
        """

    def resource_for_region(region, uri_prefix,
                            session_store):  # pragma:nocover
        """
        Get a resource for the given region.
        """


class ICredential(Interface):
    """
    An :obj:`ICredential` provides identity authentication credentials.
    """
    type_key = Attribute("The type key to look for in a JSON dictionary "
                         "representing the credential.")

    def get_session(session_store):  # pragma:nocover
        """
        Get a session corresponding to the user and tenant from the given
        session store.

        :param session_store: The store of all sessions for all tenants for
            all plugins across all regions.
        :type session_store: :class:`mimic.session.SessionStore`

        :return: A session corresponding to the user and plugin data for a
            single tenant for all plugins across all regions.
        :rtype: :class:`mimic.session.SessionStore`
        """



"""
Canned response for fastly
"""
import random
import string
import uuid


class FastlyResponse(object):

    """
    Canned response for fastly.
    See docs here https://docs.fastly.com/api/config
    """

    fastly_cache = {}

    def get_current_customer(self):
        """
        Returns the current customer with response code 200.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.get_current_customer()
                 ("/current_customer") request.
        """
        def _random_string():
            random_string = u''.join(random.choice(
                string.ascii_uppercase + string.ascii_uppercase)
                for _ in range(20))
            return random_string

        id = _random_string()
        owner_id = _random_string()

        current_customer = {
            u'can_edit_matches': u'0',
            u'can_read_public_ip_list': u'0',
            u'can_upload_vcl': u'1',
            u'updated_at': u'2014-11-03T23:37:44+00:00',
            u'has_config_panel': u'1',
            u'has_improved_ssl_config': False,
            u'id': id,
            u'has_historical_stats': u'1',
            u'has_openstack_logging': u'0',
            u'can_configure_wordpress': u'0',
            u'has_improved_logging': u'1',
            u'readonly': '',
            u'ip_whitelist': u'0.0.0.0/0',
            u'owner_id': owner_id,
            u'phone_number': u'770-123-1749',
            u'postal_address': None,
            u'billing_ref': None,
            u'can_reset_passwords': True,
            u'has_improved_security': u'1',
            u'stripe_account': None,
            u'name': u'Poppy - Test',
            u'created_at': u'2014-11-03T23:37:43+00:00',
            u'can_stream_syslog': u'1',
            u'pricing_plan': u'developer',
            u'billing_contact_id': None,
            u'has_streaming': u'1'}
        return current_customer

    def create_service(self, url_data):
        """
        Returns POST service with response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_service()
                 ("/service") request.
        """
        data = dict((key, value[0]) for key, value in url_data)

        publish_key = uuid.uuid4().hex
        service_id = uuid.uuid4().hex
        service_name = data['name']

        self.fastly_cache[service_name] = {
            'service_details': {
                u'comment': '',
                u'locked': False,
                u'updated_at': u'2014-11-13T14:29:10+00:00',
                u'created_at': u'2014-11-13T14:29:10+00:00',
                u'testing': None,
                u'number': 1,
                u'staging': None,
                u'active': None,
                u'service_id': service_id,
                u'deleted_at': None,
                u'inherit_service_id': None,
                u'deployed': None},
            'service_name': service_name
        }
        self.fastly_cache[service_id] = self.fastly_cache[service_name]

        create_service = {
            u'comment': '',
            u'publish_key': publish_key,
            u'name': service_name,
            u'versions': [{u'comment': '', u'locked': u'0',
                           u'service': service_id,
                           u'updated_at': u'2014-11-12T18:43:21',
                           u'created_at': u'2014-11-12T18:43:21',
                           u'testing': None, u'number': u'1',
                           u'staging': None,
                           u'active': None,
                           u'service_id': service_id,
                           u'deleted_at': None,
                           u'inherit_service_id': None,
                           u'deployed': None,
                           u'backend': 0}],
            u'created_at': u'2014-11-12T18:43:21+00:00',
            u'updated_at': u'2014-11-12T18:43:21+00:00',
            u'customer_id': data['customer_id'],
            u'id': service_id}
        return create_service

    def get_service_by_name(self, service_name):
        """Returns the details of the CDN service.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.get_service_by_name()
                 ("/service/version") request.
        """
        return self.fastly_cache[service_name]

    def create_version(self, service_id):
        """
        Returns POST service with response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_version()
                 ("/service/version") request.
        """
        create_version = {
            'service_id': service_id,
            'number': 1}

        return create_version

    def create_domain(self, url_data, service_id, service_version):
        """
        Returns POST create_domain with response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_domain()
                 ("/service/<service_id>/version/<service_version>/domain")
                 request.
        """
        request_dict = dict((k, v[0]) for k, v in url_data)
        domain_name = request_dict['name']

        create_domain = {
            'comment': '',
            'service_id': service_id,
            'version': service_version,
            'name': domain_name}

        if 'domain_list' not in self.fastly_cache[service_id]:
            self.fastly_cache[service_id]['domain_list'] = []

        self.fastly_cache[service_id]['domain_list'].append(
            [create_domain, 'None', 'False'])
        return create_domain

    def check_domains(self, service_id, service_version):
        """
        Returns GET check_domains with response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.check_domain()
                 ("/service/%s/version/%d/domain/check_all")
                 request.
        """
        domain_list = self.fastly_cache[service_id]['domain_list']

        return domain_list

    def create_backend(self, url_data, service_id, service_version):
        """
        Returns create_backend response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_backend()
                 ("/service/<service_id>/version/<service_version>/backend")
                 request.
        """
        request_dict = dict((k, v[0]) for k, v in url_data)

        create_backend = {
            u'comment': '',
            u'shield': None,
            u'weight': 100,
            u'ssl_client_key': None,
            u'first_byte_timeout': 15000,
            u'auto_loadbalance': False,
            u'use_ssl': request_dict['use_ssl'],
            u'port': request_dict['port'],
            u'ssl_hostname': None,
            u'hostname': request_dict['name'],
            u'error_threshold': 0,
            u'max_conn': 20,
            u'version': service_version,
            u'ipv4': None,
            u'ipv6': None,
            u'client_cert': None,
            u'ssl_ca_cert': None,
            u'request_condition': '',
            u'healthcheck': None,
            u'address': request_dict['address'],
            u'ssl_client_cert': None,
            u'name': request_dict['name'],
            u'connect_timeout': 1000,
            u'between_bytes_timeout': 10000,
            u'service_id': service_id}

        if 'origin_list' not in self.fastly_cache[service_id]:
            self.fastly_cache[service_id]['origin_list'] = []

        self.fastly_cache[service_id]['origin_list'].append(create_backend)
        return create_backend

    def create_condition(self, url_data, service_id, service_version):
        """
        Returns create_condition response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_condition()
                 ("/service/<service_id>/version/<service_version>/condition")
                 request.
        """
        request_dict = dict((k, v[0]) for k, v in url_data)

        create_condition = {
            u"type": "REQUEST",
            u"comment": "",
            u"name": "condition",
            u"version": service_version,
            u"service_id": service_id,
            u"statement": request_dict['statement'],
            u"priority": request_dict['priority']
        }

        if 'condition_list' not in self.fastly_cache[service_id]:
            self.fastly_cache[service_id]['condition_list'] = []

        self.fastly_cache[service_id][
            'condition_list'].append(create_condition)
        return create_condition

    def create_cache_settings(self, url_data, service_id, service_version):
        """
        Returns create_cache_settings response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_cache_settings()
                 ("/service/<service_id>/version/<service_version>/cache_settings")
                 request.
        """
        request_dict = dict((k, v[0]) for k, v in url_data)

        create_cache_settings = {
            "stale_ttl": request_dict.get("stale_ttl", 0),
            "ttl": request_dict.get("ttl", 0),
            "action": request_dict.get("action", ""),
            "cache_condition": "",
            "name": "cache_setting",
            "version": service_version,
            "service_id": service_id
        }

        if 'cache_settings_list' not in self.fastly_cache[service_id]:
            self.fastly_cache[service_id]['cache_settings_list'] = []

        self.fastly_cache[service_id][
            'cache_settings_list'].append(create_cache_settings)
        return create_cache_settings

    def create_response_object(self, url_data, service_id, service_version):
        """
        Returns response_object response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_response_object()
                 ("/service/<service_id>/version/<service_version>/response_object)
                 request.
        """
        request_dict = dict((k, v[0]) for k, v in url_data)

        create_response_object = {
            "status": request_dict["status"],
            "response": request_dict["response"],
            "cache_condition": request_dict.get("cache_condition", ""),
            "request_condition": request_dict.get("request_condition", ""),
            "name": request_dict["name"],
            "version": service_version,
            "content": request_dict["content"],
            "content_type": "text/plain",
            "service_id": service_id
        }

        if 'response_object_list' not in self.fastly_cache[service_id]:
            self.fastly_cache[service_id]['response_object_list'] = []

        self.fastly_cache[service_id][
            'response_object_list'].append(create_response_object)
        return create_response_object

    def create_settings(self, url_data, service_id, service_version):
        """
        Returns settings response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.create_settings()
                 ("/service/<service_id>/version/<service_version>/settings)
                 request.
        """
        request_dict = dict((k, v[0]) for k, v in url_data)

        create_settings = {
            "service_id": service_id,
            "version": service_version,
            "general.default_ttl": request_dict.get("general.default_ttl", 0),
            "general.default_host": request_dict.get("general.default_host", "")
        }

        if 'settings_list' not in self.fastly_cache[service_id]:
            self.fastly_cache[service_id]['settings_list'] = []

        self.fastly_cache[service_id][
            'settings_list'].append(create_settings)
        return create_settings

    def list_versions(self, service_id):
        """
        Returns GET list_versions with response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.list_versions()
                 ("/service/%s/version") request.
        """
        return [self.fastly_cache[service_id]['service_details']]

    def activate_version(self, service_id, version_number):
        """
        Returns activate_version response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.activate_version()
                 ("/service/%s/version/%d/activate") request.
        """
        self.fastly_cache[service_id]['service_details']['active'] = True
        return self.fastly_cache[service_id]['service_details']

    def deactivate_version(self, service_id, version_number):
        """
        Returns deactivate_version response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.deactivate_version()
                 ("/service/%s/version/%d/deactivate") request.
        """
        self.fastly_cache[service_id]['service_details']['active'] = False
        return self.fastly_cache[service_id]['service_details']

    def get_service_details(self, service_id):
        """
        Returns get_service_details response json.
        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.get_service_details()
                 ("/service/%s/details") request.
        """
        version_details = self.fastly_cache[service_id]['service_details']
        service_details = {
            u'id': service_id,
            u'name': self.fastly_cache[service_id]['service_name'],
            u'customer_id': "hTE5dRlSBICGPJxJwCH4M",
            u'comment': "",
            u"updated_at": "2012-06-14T21:20:19+00:00",
            u"created_at": "2012-06-14T21:20:19+00:00",
            u"publish_key": "xgdbdd93h5066f8d330c276fDe00f9d293abfex7",
            u'versions': [version_details]}

        return service_details

    def delete_service(self, service_id):
        """
        Returns DELETE service with response json.

        :return: a JSON-serializable dictionary matching the format of the JSON
                 response for fastly_client.delete_service()
                 ("/service/%s") request.
        """
        service_name = self.fastly_cache[service_id]['service_name']
        del(self.fastly_cache[service_id])
        del(self.fastly_cache[service_name])

        return {'status': 'ok'}

    def get_health(self):
        """
        Returns 200 with response json.

        """
        return {'status': 'ok'}



from twisted.trial.unittest import SynchronousTestCase
from twisted.internet.task import Clock

from mimic.canned_responses import fastly
from mimic.core import MimicCore
from mimic.resource import MimicRoot
from mimic.test.helpers import json_request


class FastlyAPITests(SynchronousTestCase):

    """
    Tests for the Fastly api
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`FastlyApi` as the only plugin,
        and create a service
        """
        self.core = MimicCore(Clock(), [])
        self.root = MimicRoot(self.core).app.resource()
        self.uri = '/fastly'
        self.fastly_response = fastly.FastlyResponse()

        self.customer_id = 42
        self.service_name = 'yumyum'

        (self.response, self.service_json) = self.successResultOf(json_request(
            self, self.root, "POST",
            self.uri + '/service?customer_id={0}&name={1}'.format(
                self.customer_id, self.service_name)))

        self.service_id = self.service_json['versions'][0]['service_id']
        self.version_id = self.service_json['versions'][0]['number']

    def test_get_customer(self):
        """
        ``GET /current_customer`` against the fastly API mock returns
        JSON-serialized customer details.
        """
        (response, customer_json) = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/current_customer'))
        customer_details = self.fastly_response.get_current_customer()

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(customer_json), sorted(customer_details))

    def test_create_service(self):
        """
        ``POST /service`` against the fastly API mock stores the created
        service and returns JSON-serialized service details.
        """
        service_details = self.fastly_response.create_service(
            [('customer_id', [self.customer_id]),
             ('name', [self.service_name])])

        self.assertEqual(200, self.response.code)
        self.assertEqual(sorted(self.service_json), sorted(service_details))

        (response, json_by_service_name) = self.successResultOf(json_request(
            self, self.root, "GET",
            self.uri + '/service/search?name={0}'.format(self.service_name)))
        self.assertEqual(200, self.response.code)

        (response, json_by_service_id) = self.successResultOf(json_request(
            self, self.root, "GET",
            self.uri + '/service/{0}/details'.format(self.service_id)))
        self.assertEqual(200, self.response.code)
        self.assertEqual(sorted(json_by_service_name['service_details']),
                         sorted(json_by_service_id['versions'][0]))

    def test_create_version(self):
        """
        ``POST /service/{service_id}/version)`` creates a new version of the
        service and returns JSON-serialized service details for the specific
        version.
        """
        (response, version_json) = self.successResultOf(json_request(
            self, self.root, "POST",
            self.uri + '/service/<{0}/version'.format(self.service_id)))

        version_details = self.fastly_response.create_version(
            service_id=self.service_id)

        self.assertEqual(200, self.response.code)
        self.assertEqual(sorted(version_json), sorted(version_details))

    def test_get_service_by_name(self):
        """
        ``GET /service/search`` against the Fastly mock searches by service name
        and returns JSON-serialized service details.
        """
        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "GET",
            self.uri + '/service/search?name={0}'.format(self.service_name)))
        service_details = self.fastly_response.get_service_by_name(
            service_name=self.service_name)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(service_details))

    def test_create_domain(self):
        """
        ``POST/service/{service_id}/version/{version_id}/domain`` against Fastly
        mock associates a domain to the specific version of the Fastly service
        and returns JSON-serialized service details.
        """
        uri = self.uri + '/service/{0}/version/{1}/domain' \
            '?name=llamallama&comment=redpajama'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "POST", uri))
        domain_details = self.fastly_response.create_domain(
            url_data=[('comment', ['comment']),
                      ('name', [self.service_name])],
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(domain_details))

    def test_check_domains(self):
        """
        ``GET/service/{service_id}/version/{version_id}/domain/check_all`` against
        Fastly mock returns an array with the status of all domain DNS records
        for a service version.
        """
        uri = self.uri + '/service/{0}/version/{1}/domain' \
            '?name=llamallama&comment=redpajama'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "POST", uri))

        uri = self.uri + '/service/{0}/version/{1}/domain/check_all' \
            '?name=llamallama&comment=redpajama'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "GET", uri))
        domain_details = self.fastly_response.check_domains(
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(domain_details))

    def test_create_backend(self):
        """
        ``POST /service/{service_id}/version/{version_id}/backend`` against
        Fastly mock creates a backend(origin) for a particular service and
        version and returns JSON-serialized response.
        """
        uri = self.uri + '/service/{0}/version/{1}/backend' \
            '?name=winniepoo&address=honeytree&use_ssl=False&port=80'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "POST", uri))

        url_data = [('name', ['winniepoo']), ('address', ['honeytree']),
                    ('use_ssl', [False]), ('port', [80])]
        domain_details = self.fastly_response.create_backend(
            url_data=url_data,
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(domain_details))

    def test_create_condition(self):
        """
        ``POST /service/{service_id}/version/{version_id}/condition`` against
        Fastly mock creates a condition (rule) for a particular service and
        version and returns JSON-serialized response.
        """
        uri = self.uri + '/service/{0}/version/{1}/condition' \
            '?name=testcondition&statement=req.url~+"index.html"&priority=10'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "POST", uri))

        url_data = [
            ('name', ['testcondition']),
            ('statement', ['req.url~+"index.html"']),
            ('priority', [10])
        ]
        condition = self.fastly_response.create_condition(
            url_data=url_data,
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(condition))

    def test_create_cache_settings(self):
        """
        ``POST /service/{service_id}/version/{version_id}/cache_settings`` against
        Fastly mock creates a caching setting (rule) for a particular service and
        version and returns JSON-serialized response.
        """
        uri = self.uri + '/service/{0}/version/{1}/cache_settings' \
            '?name=testcache&stale_ttl=1000&ttl=1000&action=cache'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "POST", uri))

        url_data = [
            ('name', ['testcache']),
            ('stale_ttl', ['1000']),
            ('ttl', ['1000']),
            ('action', ['cache'])
        ]
        cache_settings = self.fastly_response.create_cache_settings(
            url_data=url_data,
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(cache_settings))

    def test_create_response_object(self):
        """
        ``POST /service/{service_id}/version/{version_id}/response_object`` against
        Fastly mock creates a response_object for a particular service and
        version and returns JSON-serialized response.
        """
        uri = self.uri + '/service/{0}/version/{1}/response_object' \
            '?status=200&response=Ok&name=testresponse&content=this+message+means+all+is+okay'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "POST", uri))

        url_data = [
            ('status', ['200']),
            ('response', ['Ok']),
            ('cache_condition', [""]),
            ('request_condition', [""]),
            ('name', ['testresponse']),
            ('content', ['this+message+means+all+is+okay']),
            ('content_type', ["text/plain"]),
            ('service_id', [self.service_id])
        ]

        response_object = self.fastly_response.create_response_object(
            url_data=url_data,
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(response_object))

    def test_create_settings(self):
        """
        ``POST /service/{service_id}/version/{version_id}/settings`` against
        Fastly mock creates a settings object for a particular service and
        version and returns JSON-serialized response.
        """
        uri = self.uri + '/service/{0}/version/{1}/settings' \
            '?general.default_ttl=4242&general.default_host=www.mydomain.com'.format(
                self.service_id, self.version_id)

        (response, json_body) = self.successResultOf(json_request(
            self, self.root, "PUT", uri))

        url_data = [
            ('general.default_ttl', ['4242']),
            ('general.default_host', ['www.mydomain.com'])
        ]
        settings = self.fastly_response.create_settings(
            url_data=url_data,
            service_id=self.service_id,
            service_version=self.version_id)

        self.assertEqual(200, response.code)
        self.assertEqual(sorted(json_body), sorted(settings))

    def test_list_versions(self):
        """
        ``GET /service/{service_id}/version)`` against Fastly mock returns all
        versions associated with the service and returns a JSON-serialized
        response.
        """
        (response, version_json) = self.successResultOf(json_request(
            self, self.root, "GET",
            self.uri + '/service/{0}/version'.format(self.service_id)))

        version_details = self.fastly_response.list_versions(
            service_id=self.service_id)
        self.assertEqual(200, response.code)
        self.assertEqual(sorted(version_json), sorted(version_details))

    def test_activate_version(self):
        """
        ``PUT /service/{service_id}/version/{version_number}/activate)`` against
        Fastly mock activates the specified version of the service and returns
        a JSON-serialized response.
        """
        (response, version_json) = self.successResultOf(json_request(
            self, self.root, "PUT",
            self.uri + '/service/{0}/version/{1}/activate'.format
                       (self.service_id, self.version_id)))

        version_details = self.fastly_response.activate_version(
            service_id=self.service_id, version_number=self.version_id)
        self.assertEqual(200, response.code)
        self.assertEqual(sorted(version_json), sorted(version_details))

    def test_deactivate_version(self):
        """
        ``PUT /service/{service_id}/version/{version_number}/deactivate)``
        against Fastly mock deactivates the specified version of the service
        and returns a JSON-serialized response.
        """
        (response, version_json) = self.successResultOf(json_request(
            self, self.root, "PUT",
            self.uri + '/service/{0}/version/{1}/deactivate'.format
                       (self.service_id, self.version_id)))

        version_details = self.fastly_response.deactivate_version(
            service_id=self.service_id, version_number=self.version_id)
        self.assertEqual(200, response.code)
        self.assertEqual(sorted(version_json), sorted(version_details))

    def test_get_service_details(self):
        """
        ``GET /service/{service_id}/details`` against Fastly mock list detailed
        information on a specified service and returns a JSON-serialized
        response.
        """
        (response, service_json) = self.successResultOf(json_request(
            self, self.root, "GET",
            self.uri + '/service/{0}/details'.format(self.service_id)))

        service_details = self.fastly_response.get_service_details(
            service_id=self.service_id)
        self.assertEqual(200, response.code)
        self.assertEqual(sorted(service_json), sorted(service_details))

    def test_delete_service(self):
        """
        ``DELETE /service/{service_id}`` against Fastly mock deletes the
        specified service and returns a JSON-serialized response.
        """
        (response, delete_json) = self.successResultOf(json_request(
            self, self.root, "DELETE",
            self.uri + '/service/{0}'.format(self.service_id)))

        self.assertEqual(200, response.code)
        self.assertEqual(delete_json, {'status': 'ok'})

    def test_health(self):
        """
        ``GET /`` against Fastly mock checks if the server is up/down
        and returns a JSON-serialized response.
        """
        (response, delete_json) = self.successResultOf(json_request(
            self, self.root, "GET", self.uri))

        self.assertEqual(200, response.code)
        self.assertEqual(delete_json, {'status': 'ok'})



from __future__ import unicode_literals

from twisted.internet.task import Clock
from twisted.trial.unittest import SynchronousTestCase

from mimic.core import MimicCore
from mimic.plugins import (nova_plugin, loadbalancer_plugin, swift_plugin,
                           queue_plugin, maas_plugin, rackconnect_v3_plugin,
                           glance_plugin, cloudfeeds_plugin)


class CoreBuildingTests(SynchronousTestCase):
    """
    Tests for creating a :class:`MimicCore` object with plugins
    """
    def test_no_uuids_if_no_plugins(self):
        """
        If there are no plugins provided to :class:`MimicCore`, there are no
        uri prefixes or entries for the tenant.
        """
        core = MimicCore(Clock(), [])
        self.assertEqual(0, len(core._uuid_to_api))
        self.assertEqual([], list(core.entries_for_tenant('any_tenant', {},
                                                          'http://mimic')))

    def test_from_plugin_includes_all_plugins(self):
        """
        Using the :func:`MimicRoot.fromPlugin` creator for a
        :class:`MimicCore`, the nova and loadbalancer plugins are included.
        """
        core = MimicCore.fromPlugins(Clock())
        plugin_apis = set((
            glance_plugin.glance,
            loadbalancer_plugin.loadbalancer,
            loadbalancer_plugin.loadbalancer_control,
            maas_plugin.maas,
            maas_plugin.maas_control,
            nova_plugin.nova,
            nova_plugin.nova_control_api,
            queue_plugin.queue,
            rackconnect_v3_plugin.rackconnect,
            swift_plugin.swift,
            cloudfeeds_plugin.cloudfeeds,
            cloudfeeds_plugin.cloudfeeds_control,
        ))
        self.assertEqual(
            plugin_apis,
            set(core._uuid_to_api.values()))
        self.assertEqual(
            len(plugin_apis),
            len(list(core.entries_for_tenant('any_tenant', {},
                                             'http://mimic'))))



# -*- test-case-name: mimic.test.test_auth -*-
"""
Defines get token, impersonation
"""
import json
import time

import attr

from twisted.python.urlpath import URLPath

from mimic.canned_responses.auth import (
    get_token,
    get_endpoints,
    format_timestamp,
    impersonator_user_role)
from mimic.canned_responses.mimic_presets import get_presets
from mimic.core import MimicCore
from mimic.model.behaviors import make_behavior_api
from mimic.model.identity import (
    APIKeyCredentials,
    ImpersonationCredentials,
    PasswordCredentials,
    TokenCredentials)
from mimic.rest.mimicapp import MimicApp
from mimic.session import NonMatchingTenantError
from mimic.util.helper import (
    invalid_resource,
    seconds_to_timestamp)

from mimic.model.behaviors import (
    BehaviorRegistryCollection,
    Criterion,
    EventDescription,
    regexp_predicate
)

authentication = EventDescription()
"""
Event refers to authenticating against Identity using a username/password,
username/api-key, token, or getting an impersonation token.
"""


@authentication.declare_criterion("username")
def username_criterion(value):
    """
    Return a Criterion which matches the given regular expression string
    against the ``"username"`` attribute.
    """
    return Criterion(name='username', predicate=regexp_predicate(value))


@authentication.declare_criterion("tenant_id")
def tenant_id_criterion(value):
    """
    Return a Criterion which matches the given regular expression string
    against the ``"tenant_Id"`` attribute.
    """
    return Criterion(name='tenant_id', predicate=regexp_predicate(value))


@authentication.declare_default_behavior
def default_authentication_behavior(core, http_request, credentials):
    """
    Default behavior in response to a server creation.  This will create
    a session for the tenant if one does not already exist, and return
    the auth token for that session.  In the case of
    :class:`PasswordCredentials`, :class:`ApiKeyCredentials`, or
    :class:`TokenCredentials`, also returns the service catalog.

    :param core: An instance of :class:`mimic.core.MimicCore`
    :param http_request: A twisted http request/response object
    :param credentials: An `mimic.model.identity.ICredentials` provider

    Handles setting the response code and also
    :return: The response body for a default authentication request.
    """
    try:
        session = credentials.get_session(core.sessions)
    except NonMatchingTenantError as e:
        http_request.setResponseCode(401)
        if type(credentials) == TokenCredentials:
            message = ("Token doesn't belong to Tenant with Id/Name: "
                       "'{0}'".format(e.desired_tenant))
        else:
            message = ("Tenant with Name/Id: '{0}' is not valid for "
                       "User '{1}' (id: '{2}')".format(
                           e.desired_tenant,
                           e.session.username,
                           e.session.user_id))

        return json.dumps({
            "unauthorized": {
                "code": 401,
                "message": message
            }
        })
    else:
        if type(credentials) == ImpersonationCredentials:
            return json.dumps({"access": {
                "token": {"id": credentials.impersonated_token,
                          "expires": format_timestamp(session.expires)}
            }})

        http_request.setResponseCode(200)
        prefix_map = {
            # map of entry to URI prefix for that entry
        }

        def lookup(entry):
            return prefix_map[entry]
        result = get_token(
            session.tenant_id,
            entry_generator=lambda tenant_id:
            list(core.entries_for_tenant(
                 session.tenant_id, prefix_map,
                 base_uri_from_request(http_request))),
            prefix_for_endpoint=lookup,
            response_token=session.token,
            response_user_id=session.user_id,
            response_user_name=session.username,
        )
        return json.dumps(result)


@authentication.declare_behavior_creator("fail")
def authenticate_failure_behavior(parameters):
    """
    Create a failing behavior for authentication.

    Takes three parameters:

    ``"code"``, an integer describing the HTTP response code, and
    ``"message"``, a string describing a textual message.
    ``"type"``, a string representing what type of error message it is

    If ``type`` is "string", the message is just returned as the string body.
    Otherwise, the following JSON body will be synthesized (as per the
    canonical Nova error format):

    ```
    {
        <type>: {
            "message": <message>,
            "code": <code>
        }
    }

    The default type is unauthorized, the default code is 401, and the
    default message is
    "Unable to authenticate user with credentials provided."
    """
    def _fail(core, http_request, credentials):
        status_code = parameters.get("code", 401)
        http_request.setResponseCode(status_code)

        failure_type = parameters.get("type", "unauthorized")
        failure_message = parameters.get(
            "message",
            "Unable to authenticate user with credentials provided.")

        if failure_type == "string":
            return failure_message
        else:
            return json.dumps({
                failure_type: {
                    "message": failure_message,
                    "code": status_code
                }
            })

    return _fail


@attr.s(hash=False)
class AuthApi(object):
    """
    Rest endpoints for mocked Auth api.

    :ivar core: an instance of :class:`mimic.core.MimicCore`
    :ivar registry_collection: an instance of
        :class:`mimic.model.behaviors.BehaviorRegistryCollection`
    """
    core = attr.ib(validator=attr.validators.instance_of(MimicCore))
    registry_collection = attr.ib(
        validator=attr.validators.instance_of(BehaviorRegistryCollection))
    app = MimicApp()

    @app.route('/v2.0/tokens', methods=['POST'])
    def get_token_and_service_catalog(self, request):
        """
        Return a service catalog consisting of all plugin endpoints and an api
        token.
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            pass
        else:
            for cred_type in (PasswordCredentials, APIKeyCredentials,
                              TokenCredentials):
                if cred_type.type_key in content['auth']:
                    try:
                        cred = cred_type.from_json(content)
                    except (KeyError, TypeError):
                        pass
                    else:
                        registry = self.registry_collection.registry_by_event(
                            authentication)
                        behavior = registry.behavior_for_attributes(
                            attr.asdict(cred))
                        return behavior(self.core, request, cred)

        request.setResponseCode(400)
        return json.dumps(invalid_resource("Invalid JSON request body"))

    @app.route('/v1.1/mosso/<string:tenant_id>', methods=['GET'])
    def get_username(self, request, tenant_id):
        """
        Returns response with random usernames.
        """
        request.setResponseCode(301)
        session = self.core.sessions.session_for_tenant_id(tenant_id)
        return json.dumps(dict(user=dict(id=session.username)))

    @app.route('/v2.0/users', methods=['GET'])
    def get_users_details(self, request):
        """
        Returns response with  detailed account information about each user
        including email, name, user ID, account configuration and status
        information.
        """
        username = request.args.get("name")
        session = self.core.sessions.session_for_username_password(
            username[0], "test")
        return json.dumps(dict(user={
            "RAX-AUTH:domainId": session.tenant_id,
            "id": session.user_id,
            "enabled": True,
            "username": session.username,
            "email": "thisisrandom@email.com",
            "RAX-AUTH:defaultRegion": "ORD",
            "created": seconds_to_timestamp(time.time()),
            "updated": seconds_to_timestamp(time.time())
        }))

    @app.route('/v2.0/users/<string:user_id>/OS-KSADM/credentials/RAX-KSKEY:apiKeyCredentials',
               methods=['GET'])
    def rax_kskey_apikeycredentials(self, request, user_id):
        """
        Support, such as it is, for the apiKeysCredentials call.
        """
        if user_id in self.core.sessions._userid_to_session:
            username = self.core.sessions._userid_to_session[user_id].username.decode('ascii')
            apikey = '7fc56270e7a70fa81a5935b72eacbe29'  # echo -n A | md5sum
            return json.dumps({'RAX-KSKEY:apiKeyCredentials': {'username': username,
                                                               'apiKey': apikey}})
        else:
            request.setResponseCode(404)
            return json.dumps({'itemNotFound':
                              {'code': 404, 'message': 'User ' + user_id + ' not found'}})

    @app.route('/v2.0/RAX-AUTH/impersonation-tokens', methods=['POST'])
    def get_impersonation_token(self, request):
        """
        Return a token id with expiration.
        """
        request.setResponseCode(200)
        try:
            content = json.loads(request.content.read())
        except ValueError:
            request.setResponseCode(400)
            return json.dumps(invalid_resource("Invalid JSON request body"))

        cred = ImpersonationCredentials.from_json(
            content, request.getHeader("x-auth-token"))
        registry = self.registry_collection.registry_by_event(authentication)
        behavior = registry.behavior_for_attributes({
            "token": cred.impersonator_token,
            "username": cred.impersonated_username
        })
        return behavior(self.core, request, cred)

    @app.route('/v2.0/tokens/<string:token_id>', methods=['GET'])
    def validate_token(self, request, token_id):
        """
        Creates a new session for the given tenant_id and token_id
        and always returns response code 200.
        Docs: http://developer.openstack.org/api-ref-identity-v2.html#admin-tokens
        """
        request.setResponseCode(200)
        tenant_id = request.args.get('belongsTo')
        if tenant_id is not None:
            tenant_id = tenant_id[0]
        session = self.core.sessions.session_for_tenant_id(tenant_id, token_id)
        response = get_token(
            session.tenant_id,
            response_token=session.token,
            response_user_id=session.user_id,
            response_user_name=session.username,
        )
        if session.impersonator_session_for_token(token_id) is not None:
            impersonator_session = session.impersonator_session_for_token(token_id)
            response["access"]["RAX-AUTH:impersonator"] = impersonator_user_role(
                impersonator_session.user_id,
                impersonator_session.username)

        if token_id in get_presets["identity"]["token_fail_to_auth"]:
            request.setResponseCode(401)
            return json.dumps({'itemNotFound':
                              {'code': 401, 'message': 'Invalid auth token'}})

        imp_token = get_presets["identity"]["maas_admin_roles"]
        racker_token = get_presets["identity"]["racker_token"]
        if token_id in imp_token:
            response["access"]["RAX-AUTH:impersonator"] = {
                "id": response["access"]["user"]["id"],
                "name": response["access"]["user"]["name"],
                "roles": [{"id": "123",
                           "name": "monitoring:service-admin"},
                          {"id": "234",
                           "name": "object-store:admin"}]}
        if token_id in racker_token:
            response["access"]["RAX-AUTH:impersonator"] = {
                "id": response["access"]["user"]["id"],
                "name": response["access"]["user"]["name"],
                "roles": [{"id": "9",
                           "name": "Racker"}]}
        if tenant_id in get_presets["identity"]["observer_role"]:
            response["access"]["user"]["roles"] = [
                {"id": "observer",
                 "description": "Global Observer Role.",
                 "name": "observer"}]
        if tenant_id in get_presets["identity"]["creator_role"]:
            response["access"]["user"]["roles"] = [
                {"id": "creator",
                 "description": "Global Creator Role.",
                 "name": "creator"}]
        if tenant_id in get_presets["identity"]["admin_role"]:
            response["access"]["user"]["roles"] = [
                {"id": "admin",
                 "description": "Global Admin Role.",
                 "name": "admin"},
                {"id": "observer",
                 "description": "Global Observer Role.",
                 "name": "observer"}]

        if token_id in get_presets["identity"]["non_dedicated_observer"]:
                response["access"]["token"]["tenant"] = {
                    "id": "135790",
                    "name": "135790",
                }
                response["access"]["user"] = {
                    "name": "OneTwo",
                    "roles": [{"id": "1",
                               "name": "monitoring:observer",
                               "description": "Monitoring Observer"}]
                }

        if token_id in get_presets["identity"]["non_dedicated_admin"]:
                response["access"]["token"]["tenant"] = {
                    "id": "135790",
                    "name": "135790",
                }
                response["access"]["user"] = {
                    "name": "ThreeFour",
                    "roles": [{"id": "1",
                               "name": "monitoring:admin",
                               "description": "Monitoring Admin"},
                              {"id": "2",
                               "name": "admin",
                               "description": "Admin"}]
                }

        if token_id in get_presets["identity"]["dedicated_full_device_permission_holder"]:
                response["access"]["token"]["tenant"] = {
                    "id": "hybrid:123456",
                    "name": "hybrid:123456",
                }
                response["access"]["user"] = {
                    "id": "12",
                    "name": "HybridOneTwo",
                    "roles": [{"id": "1",
                               "name": "monitoring:observer",
                               "description": "Monitoring Observer"},
                              {"id": "3",
                               "name": "hybridRole",
                               "description": "Hybrid Admin",
                               "tenantId": "hybrid:123456"}],
                    "RAX-AUTH:contactId": "12"
                }

        if token_id in get_presets["identity"]["dedicated_account_permission_holder"]:
                response["access"]["token"]["tenant"] = {
                    "id": "hybrid:123456",
                    "name": "hybrid:123456",
                }
                response["access"]["user"] = {
                    "id": "34",
                    "name": "HybridThreeFour",
                    "roles": [{"id": "1",
                               "name": "monitoring:creator",
                               "description": "Monitoring Creator"},
                              {"id": "2",
                               "name": "creator",
                               "description": "Creator"}],
                    "RAX-AUTH:contactId": "34"
                }

        if token_id in get_presets["identity"]["dedicated_limited_device_permission_holder"]:
                response["access"]["token"]["tenant"] = {
                    "id": "hybrid:123456",
                    "name": "hybrid:123456",
                }
                response["access"]["user"] = {
                    "id": "56",
                    "name": "HybridFiveSix",
                    "roles": [{"id": "1",
                               "name": "monitoring:observer",
                               "description": "Monitoring Observer"},
                              {"id": "2",
                               "name": "observer",
                               "description": "Observer"}],
                    "RAX-AUTH:contactId": "56"
                }

        if token_id in get_presets["identity"]["dedicated_other_account_observer"]:
                response["access"]["token"]["tenant"] = {
                    "id": "hybrid:654321",
                    "name": "hybrid:654321",
                }
                response["access"]["user"] = {
                    "id": "78",
                    "name": "HybridSevenEight",
                    "roles": [{"id": "1",
                               "name": "monitoring:observer",
                               "description": "Observer"},
                              {"id": "2",
                               "name": "observer",
                               "description": "Observer"}],
                    "RAX-AUTH:contactId": "78"
                }

        if token_id in get_presets["identity"]["dedicated_other_account_admin"]:
                response["access"]["token"]["tenant"] = {
                    "id": "hybrid:654321",
                    "name": "hybrid:654321",
                }
                response["access"]["user"] = {
                    "id": "90",
                    "name": "HybridNineZero",
                    "roles": [{"id": "1",
                               "name": "monitoring:admin",
                               "description": "Admin"},
                              {"id": "2",
                               "name": "admin",
                               "description": "Admin"}],
                    "RAX-AUTH:contactId": "90"
                }

        return json.dumps(response)

    @app.route('/v2.0/tokens/<string:token_id>/endpoints', methods=['GET'])
    def get_endpoints_for_token(self, request, token_id):
        """
        Return a service catalog consisting of nova and load balancer mocked
        endpoints.
        """
        # FIXME: TEST
        request.setResponseCode(200)
        prefix_map = {}
        session = self.core.sessions.session_for_token(token_id)
        return json.dumps(get_endpoints(
            session.tenant_id,
            entry_generator=lambda tenant_id: list(
                self.core.entries_for_tenant(tenant_id, prefix_map,
                                             base_uri_from_request(request))),
            prefix_for_endpoint=prefix_map.get)
        )


def base_uri_from_request(request):
    """
    Given a request, return the base URI of the request

    :param request: a twisted HTTP request
    :type request: :class:`twisted.web.http.Request`

    :return: the base uri the request was trying to access
    :rtype: ``str``
    """
    return str(URLPath.fromRequest(request).click('/'))


AuthControlApiBehaviors = make_behavior_api({'auth': authentication})
"""
Handlers for CRUD operations on authentication behaviors.

:ivar registry_collection: an instance of
    :class:`mimic.model.behaviors.BehaviorRegistryCollection`
"""



"""
Unit tests for :mod:`mimic.util`
"""
from twisted.trial.unittest import SynchronousTestCase
from twisted.web.resource import Resource

from mimic.util import helper
from mimic.test.helpers import request


class HelperTests(SynchronousTestCase):
    """
    Tests for :mod:`mimic.util.helper`
    """

    matches = [(0, "1970-01-01T00:00:00.000000Z"),
               (1.5, "1970-01-01T00:00:01.500000Z"),
               (121.4005, "1970-01-01T00:02:01.400500Z")]

    def _validate_ipv4_address(self, address, *prefixes):
        nums = [int(x) for x in address.split('.')]
        self.assertEqual(4, len(nums))
        self.assertTrue(all(x >= 0 and x < 256 for x in nums))
        self.assertEqual(prefixes[:len(nums)], tuple(nums[:len(prefixes)]))

    def test_random_ipv4_completely_random(self):
        """
        A completely random IP address is generated if prefixes are not
        provided.
        """
        self._validate_ipv4_address(helper.random_ipv4())

    def test_random_ipv4_with_prefixes(self):
        """
        Random IP addresses can be generated with pre-determined prefixes.
        """
        prefixes = []
        for i in range(1, 5):
            prefixes.append(i)
            self._validate_ipv4_address(helper.random_ipv4(*prefixes),
                                        *prefixes)

    def test_random_hex_generator(self):
        """
        A completely random and unique hex encoded data is generated.
        """
        self.assertNotEqual(helper.random_hex_generator(3),
                            helper.random_hex_generator(3))
        self.assertEqual(len(helper.random_hex_generator(4)), 8)

    def test_seconds_to_timestamp_default_timestamp(self):
        """
        :func:`helper.seconds_to_timestamp` returns a timestamp matching
        the seconds since the epoch given.  The timestamp conforms to with the
        default format string of ``%Y-%m-%dT%H:%M:%S.%fZ`` if no format
        string is provided.
        """
        for seconds, timestamp in self.matches:
            self.assertEqual(timestamp, helper.seconds_to_timestamp(seconds))

    def test_seconds_to_timestamp_provided_timestamp(self):
        """
        :func:`helper.seconds_to_timestamp` uses the provided timestamp format
        to format the seconds.
        """
        formats = [("%m-%d-%Y %H:%M:%S", "01-01-1970 00:00:00"),
                   ("%Y-%m-%d", "1970-01-01"),
                   ("%H %M %S (%f)", "00 00 00 (000000)")]
        for fmt, timestamp in formats:
            self.assertEqual(timestamp,
                             helper.seconds_to_timestamp(0, fmt))

    def test_timestamp_to_seconds(self):
        """
        :func:`helper.timestamp_to_seconds` returns a seconds since EPOCH
        matching the corresponding timestamp given in ISO8601 format
        """

        local_matches = [(0, "1970-01-01"), (0, "1970-01-01T00"),
                         (0, "1970-01-01T00:00"), (1800, "1970-01-01T00-00:30")]
        for seconds, timestamp in self.matches + local_matches:
            self.assertEqual(seconds, helper.timestamp_to_seconds(timestamp))


class TestHelperTests(SynchronousTestCase):
    """
    Tests for :obj:`mimic.test.helpers`.
    """

    def test_unicode_body(self):
        """
        If :obj:`request` is given a unicode request body, the deferred
        synchronously fails so that the caller can immediately tell something
        is wrong.
        """
        self.failureResultOf(
            request(self, Resource(), b"POST", b"", u"not bytes")
        )


class TestRandomString(SynchronousTestCase):
    """
    Tests for random string generation.
    """

    def test_length(self):
        """
        The random string you generate should have the length you specify.
        """
        for l in range(100):
            self.assertEqual(len(helper.random_string(l)), l)

    def test_selectable(self):
        """
        When passing a custom selectable, the results should derive only from
        the characters you provide.
        """
        desired_chars = "02468"
        for iteration in xrange(100):
            a_string = helper.random_string(1024, selectable=desired_chars)
            for char in a_string:
                self.assertTrue(char in desired_chars)



"""
Model objects for the Valkyrie mimic.
"""
from characteristic import attributes, Attribute
from json import dumps

from mimic.util.helper import random_hex_generator


class AccountContactPermission(object):
    """
    An intersection object representing a certain contact's permissions on a certain account's items
    """

    permission_type_map = {
        6: "view_domain",
        4: "view_billing",
        14: "admin_product",  # device.admin
        10: "manage_users",
        8: "manage_certificates",
        19: "edit_firewall_config",
        2: "edit_ticket",
        18: "view_firewall_config",
        15: "account_admin",  # account.admin
        7: "edit_domain",
        13: "edit_product",   # device.admin
        3: "view_community",
        17: "view_reports",
        16: "move_manager",
        12: "view_product",   # device.observer
        11: "manage_contact",
        9: "upgrade_account",
        5: "edit_billing",
        1: "view_ticket"}

    item_type_map = {
        1: "accounts",
        2: "devices"}

    def __init__(self, account_number, contact_id, permission_type, item_id, item_type_id):
        """
        Constructor
        """
        self.account_number = account_number
        self.contact_id = contact_id
        self.permission_type = permission_type
        self.permission_name = self.permission_type_map.get(self.permission_type, "unknown")
        self.item_id = item_id
        self.item_type_id = item_type_id
        self.item_type_name = self.item_type_map.get(self.item_type_id, "unknown")

    def json(self):
        """
        Create a JSON representation of self
        """
        return {
            "account_number": self.account_number,
            "contact_id": self.contact_id,
            "permission_type": self.permission_type,
            "permission_name": self.permission_name,
            "item_id": self.item_id,
            "item_type_id": self.item_type_id,
            "item_type_name": self.item_type_name
        }


@attributes([Attribute("valkyrie_store", default_factory=list)])
class ValkyrieStore(object):
    """

    Extremely barebones Valkyrie backing store with some direct, static permissions.

    No create or delete permissions endpoints are implemented.
    No logic for determining effective permissions from indirect permissions is present.

    A GET on the following URI, for example, should always return four effective permissions:

        http://localhost:8900/valkyrie/v2/account/123456/permissions/contacts/devices/by_contact/12/effective

    ...while a GET on this URI should return one:

        http://localhost:8900/valkyrie/v2/account/123456/permissions/contacts/devices/by_contact/56/effective

    """

    permissions = []
    # Arguments are: account, contact, (direct) permission, item, item_type (1=account or 2=device)
    permissions.append(AccountContactPermission(123456, 12, 12, 256, 2))
    permissions.append(AccountContactPermission(123456, 12, 12, 4096, 2))
    permissions.append(AccountContactPermission(123456, 12, 13, 16384, 2))
    permissions.append(AccountContactPermission(123456, 12, 14, 65536, 2))
    permissions.append(AccountContactPermission(123456, 34, 15, 123456, 1))
    permissions.append(AccountContactPermission(123456, 56, 12, 256, 2))

    permissions.append(AccountContactPermission(654321, 78, 14, 262144, 2))
    permissions.append(AccountContactPermission(654321, 90, 12, 1048576, 2))
    permissions.append(AccountContactPermission(654321, 90, 15, 654321, 1))

    def create_token(self, request):
        """
        Create an auth token without even interrogating the POSTed credential data
        """
        request.setResponseCode(200)
        token = {"X-Auth-Token": str(random_hex_generator(16))}
        return dumps(token)

    def get_permissions(self, request, account_number, contact_id, item_type):
        """
        Retrieve the permissions (if any) belonging to the given account,
        contact, and item type (item_type=1 -> accounts, item_type=2 -> devices
        """
        pm = [p for p in self.permissions if (p.account_number == account_number and
                                              p.contact_id == contact_id and
                                              (item_type is None or p.item_type_id == item_type))]

        response_message = {"contact_permissions": []}
        for p in pm:
            response_message['contact_permissions'].append(p.json())

        return dumps(response_message)



# -*- test-case-name: mimic.test.test_loadbalancer -*-
"""
Defines add node and delete node from load balancers
"""
import json
from uuid import uuid4
from six import string_types, text_type
from zope.interface import implementer
from twisted.plugin import IPlugin
from mimic.rest.mimicapp import MimicApp
from mimic.imimic import IAPIMock
from mimic.catalog import Entry
from mimic.catalog import Endpoint

from mimic.model.clb_errors import invalid_json_schema
from mimic.model.clb_objects import (
    GlobalCLBCollections, BadKeysError, BadValueError
)
from random import randrange

from mimic.util.helper import invalid_resource, json_dump
from characteristic import attributes


@implementer(IAPIMock, IPlugin)
class LoadBalancerApi(object):
    """
    This class registers the load balancer API in the service catalog.
    """
    def __init__(self, regions=["ORD"]):
        """
        Create an API with the specified regions.
        """
        self._regions = regions

    def catalog_entries(self, tenant_id):
        """
        Cloud load balancer entries.
        """
        return [
            Entry(tenant_id, "rax:load-balancer", "cloudLoadBalancers",
                  [
                      Endpoint(tenant_id, region, text_type(uuid4()),
                               prefix="v2")
                      for region in self._regions
                  ])
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        lb_region = LoadBalancerRegion(self, uri_prefix, session_store,
                                       region)
        return lb_region.app.resource()

    def _get_session(self, session_store, tenant_id):
        """
        Retrieve or create a new LoadBalancer session from a given tenant identifier
        and :obj:`SessionStore`.

        For use with ``data_for_api``.

        Temporary hack; see this issue
        https://github.com/rackerlabs/mimic/issues/158
        """
        return (
            session_store.session_for_tenant_id(tenant_id)
            .data_for_api(self, lambda: GlobalCLBCollections(
                clock=session_store.clock
            ))
        )


@implementer(IAPIMock, IPlugin)
@attributes(["lb_api"])
class LoadBalancerControlApi(object):
    """
    This class registers the load balancer controller API in the service
    catalog.
    """
    def catalog_entries(self, tenant_id):
        """
        Cloud load balancer controller endpoints.
        """
        return [
            Entry(
                tenant_id, "rax:load-balancer", "cloudLoadBalancerControl",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()), prefix="v2")
                    for region in self.lb_api._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        lbc_region = LoadBalancerControlRegion(api_mock=self, uri_prefix=uri_prefix,
                                               session_store=session_store, region=region)
        return lbc_region.app.resource()


@attributes(["api_mock", "uri_prefix", "session_store", "region"])
class LoadBalancerControlRegion(object):
    """
    Klein routes for load balancer's control API within a particular region.
    """

    app = MimicApp()

    def _collection_from_tenant(self, tenant_id):
        """
        Retrieve the server collection for this region for the given tenant.
        """
        return (self.api_mock.lb_api._get_session(self.session_store, tenant_id)
                .collection_for_region(self.region))

    @app.route(
        '/v2/<string:tenant_id>/loadbalancer/<int:clb_id>/attributes',
        methods=['PATCH']
    )
    def set_attributes(self, request, tenant_id, clb_id):
        """
        Alters the supported attributes of the CLB to supported values.  To
        return things back to normal, you'll first need to list the CLB to get
        any original values yourself.
        """
        regional_lbs = self._collection_from_tenant(tenant_id)
        if not regional_lbs.lb_in_region(clb_id):
            request.setResponseCode(404)
            return json.dumps({
                "message": "Tenant {0} doesn't own load balancer {1}".format(
                    tenant_id, clb_id
                ),
                "code": 404,
            })

        try:
            content = json.loads(request.content.read())
        except ValueError:
            request.setResponseCode(400)
            return json.dumps(invalid_resource("Invalid JSON request body"))

        try:
            regional_lbs.set_attributes(clb_id, content)
        except BadKeysError, bke:
            request.setResponseCode(400)
            return json.dumps({
                "message": str(bke),
                "code": 400,
            })
        except BadValueError, bve:
            request.setResponseCode(400)
            return json.dumps({
                "message": str(bve),
                "code": 400,
            })
        else:
            request.setResponseCode(204)
            return b''


class LoadBalancerRegion(object):
    """
    Klein routes for load balancer API methods within a particular region.
    """

    app = MimicApp()

    def __init__(self, api_mock, uri_prefix, session_store, region_name):
        """
        Fetches the load balancer id for a failure, invalid scenarios and
        the count on the number of time 422 should be returned on add node.
        """
        self.uri_prefix = uri_prefix
        self.region_name = region_name
        self._api_mock = api_mock
        self._session_store = session_store

    def session(self, tenant_id):
        """
        Gets a session for a particular tenant, creating one if there isn't
        one.
        """
        tenant_session = self._session_store.session_for_tenant_id(tenant_id)
        clb_global_collection = tenant_session.data_for_api(
            self._api_mock,
            lambda: GlobalCLBCollections(
                clock=self._session_store.clock))
        clb_region_collection = clb_global_collection.collection_for_region(
            self.region_name)
        return clb_region_collection

    @app.route('/v2/<string:tenant_id>/loadbalancers', methods=['POST'])
    def add_load_balancer(self, request, tenant_id):
        """
        Creates a load balancer and adds it to the load balancer store.
        Returns the newly created load balancer with response code 202
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            request.setResponseCode(400)
            return json.dumps(invalid_resource("Invalid JSON request body"))

        lb_id = randrange(99999)
        response_data = self.session(tenant_id).add_load_balancer(
            content['loadBalancer'], lb_id
        )
        request.setResponseCode(response_data[1])
        return json.dumps(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>', methods=['GET'])
    def get_load_balancers(self, request, tenant_id, lb_id):
        """
        Returns a list of all load balancers created using mimic with response code 200
        """
        response_data = self.session(tenant_id).get_load_balancers(lb_id)
        request.setResponseCode(response_data[1])
        return json.dumps(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers', methods=['GET'])
    def list_load_balancers(self, request, tenant_id):
        """
        Returns a list of all load balancers created using mimic with response code 200
        """
        response_data = self.session(tenant_id).list_load_balancers()
        request.setResponseCode(response_data[1])
        return json.dumps(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>', methods=['DELETE'])
    def delete_load_balancer(self, request, tenant_id, lb_id):
        """
        Creates a load balancer and adds it to the load balancer store.
        Returns the newly created load balancer with response code 200
        """
        response_data = self.session(tenant_id).del_load_balancer(lb_id)
        request.setResponseCode(response_data[1])
        return json_dump(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes', methods=['POST'])
    def add_node_to_load_balancer(self, request, tenant_id, lb_id):
        """
        Return a successful add node response with code 200
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            request.setResponseCode(400)
            return json.dumps(invalid_resource("Invalid JSON request body"))

        node_list = content['nodes']
        response_data = self.session(tenant_id).add_node(node_list, lb_id)
        request.setResponseCode(response_data[1])
        return json.dumps(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes/<int:node_id>',
               methods=['GET'])
    def get_node(self, request, tenant_id, lb_id, node_id):
        """
        Returns a 200 response code and a particular node on the load balancer
        """
        body, code = self.session(tenant_id).get_node(lb_id, node_id)
        request.setResponseCode(code)
        return json.dumps(body)

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes/<int:node_id>.atom',
               methods=['GET'])
    def get_node_feed(self, request, tenant_id, lb_id, node_id):
        """
        Returns a 200 response code and node's feed on the load balancer
        """
        body, code = self.session(tenant_id).get_node_feed(lb_id, node_id)
        request.setResponseCode(code)
        request.setHeader("Content-Type", "application/atom+xml")
        return body

    @app.route(
        '/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes/<int:node_id>',
        methods=['PUT'])
    def update_node(self, request, tenant_id, lb_id, node_id):
        """
        Return a 202 response code to updating a node, if successful.
        """
        try:
            content = json.loads(request.content.read())
            assert isinstance(content, dict) and content.keys() == ["node"]
            content = content["node"]
            assert isinstance(content, dict)
        except (ValueError, AssertionError):
            resp_body, resp_code = invalid_json_schema()
        else:
            resp_body, resp_code = self.session(tenant_id).update_node(
                lb_id, node_id, content
            )
        request.setResponseCode(resp_code)
        if isinstance(resp_body, string_types):
            return resp_body
        return json.dumps(resp_body)

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes/<int:node_id>',
               methods=['DELETE'])
    def delete_node_from_load_balancer(self, request, tenant_id, lb_id, node_id):
        """
        Returns a 204 response code, for any load balancer created using the mocks
        """
        response_data = self.session(tenant_id).delete_node(lb_id, node_id)
        request.setResponseCode(response_data[1])
        return json.dumps(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes',
               methods=['DELETE'])
    def delete_nodes_from_load_balancer(self, request, tenant_id, lb_id):
        """
        Deletes multiple nodes from a LB.
        """
        node_ids = map(int, request.args.get('id', []))
        response_data = self.session(tenant_id).delete_nodes(lb_id, node_ids)
        request.setResponseCode(response_data[1])
        return json_dump(response_data[0])

    @app.route('/v2/<string:tenant_id>/loadbalancers/<int:lb_id>/nodes',
               methods=['GET'])
    def list_nodes_for_load_balancer(self, request, tenant_id, lb_id):
        """
        Returns a 200 response code and list of nodes on the load balancer
        """
        response_data = self.session(tenant_id).list_nodes(lb_id)
        request.setResponseCode(response_data[1])
        return json.dumps(response_data[0])



"""
Model objects for the flavors.
"""

from characteristic import attributes, Attribute
from json import dumps
from mimic.model.flavors import (
    RackspaceStandardFlavor, RackspaceComputeFlavor, RackspaceMemoryFlavor,
    RackspaceOnMetalFlavor, RackspaceIOFlavor, RackspaceGeneralFlavor,
    RackspacePerformance1Flavor, RackspacePerformance2Flavor)

from mimic.model.nova_objects import not_found


@attributes(
    ["tenant_id", "region_name", "clock",
     Attribute("flavors_store", default_factory=list)]
)
class RegionalFlavorCollection(object):
    """
    A collection of flavors, in a given region, for a given tenant.
    """
    def flavor_by_id(self, flavor_id):
        """
        Retrieve a :obj:`Flavor` object by its ID.
        """
        self._create_flavors_list()
        for flavor in self.flavors_store:
            if flavor.flavor_id == flavor_id:
                return flavor

    def _create_flavors_list(self):
        """
        Generates the data for each flavor in each flavor class
        """
        rackspace_flavors = [RackspaceStandardFlavor, RackspaceComputeFlavor,
                             RackspacePerformance1Flavor, RackspaceOnMetalFlavor,
                             RackspacePerformance2Flavor, RackspaceMemoryFlavor,
                             RackspaceIOFlavor, RackspaceGeneralFlavor]

        if len(self.flavors_store) < 1:
            for flavor_class in rackspace_flavors:
                for flavor, flavor_spec in flavor_class.flavors.iteritems():
                    flavor_name = flavor
                    flavor_id = flavor_spec['id']
                    ram = flavor_spec['ram']
                    vcpus = flavor_spec['vcpus']
                    network = flavor_spec['rxtx_factor']
                    disk = flavor_spec['disk']
                    tenant_id = self.tenant_id
                    flavor = flavor_class(flavor_id=flavor_id, tenant_id=tenant_id,
                                          name=flavor_name, ram=ram, vcpus=vcpus,
                                          rxtx=network, disk=disk)
                    self.flavors_store.append(flavor)

    def list_flavors(self, include_details, absolutize_url):
        """
        Return a list of flavors with details.
        """
        self._create_flavors_list()
        flavors = []
        for flavor in self.flavors_store:
            if self.region_name != "IAD" and isinstance(flavor, RackspaceOnMetalFlavor):
                continue
            if include_details:
                flavors.append(flavor.detailed_json(absolutize_url))
            else:
                flavors.append(flavor.brief_json(absolutize_url))
        result = {"flavors": flavors}

        return dumps(result)

    def get_flavor(self, http_get_request, flavor_id, absolutize_url):
        """
        Return a flavor object if one exists from the list `/flavors` api,
        If the `flavor_id` is not found in the flavor store,
        then will return 404.
        """
        flavor = self.flavor_by_id(flavor_id)
        if flavor is None:
            return dumps(not_found("The resource could not be found.",
                                   http_get_request))
        return dumps({"flavor": flavor.detailed_json(absolutize_url)})


@attributes(["tenant_id", "clock",
             Attribute("regional_collections", default_factory=dict)])
class GlobalFlavorCollection(object):
    """
    A :obj:`GlobalFlavorCollection` is a set of all the
    :obj:`RegionalFlavorCollection` objects owned by a given tenant.  In other
    words, all the flavor objects that a single tenant owns globally.
    """

    def collection_for_region(self, region_name):
        """
        Get a :obj:`RegionalFlavorCollection` for the region identified by the
        given name.
        """
        if region_name not in self.regional_collections:
            self.regional_collections[region_name] = (
                RegionalFlavorCollection(tenant_id=self.tenant_id, region_name=region_name,
                                         clock=self.clock))
        return self.regional_collections[region_name]



"""
Emulate the behavior of Rackspace Cloud Feeds service.
Please refer to
http://docs.rackspace.com/cloud-feeds/api/v1.0/feeds-devguide/content/overview.html
for more details.
"""

import attr
from six import string_types


@attr.s
class CloudFeedsProduct(object):
    """
    Models a single CloudFeed product endpoint and its respective
    functionality.
    """
    title = attr.ib(validator=attr.validators.instance_of(string_types))
    href = attr.ib(validator=attr.validators.instance_of(string_types))

    events = attr.ib(default=attr.Factory(list))

    def post(self, event):
        """
        Post an event to the product queue.
        """
        self.events.append(event)


@attr.s
class CloudFeeds(object):
    """
    Models CloudFeeds support at the plugin-level.
    """
    tenant_id = attr.ib(validator=attr.validators.instance_of(string_types))
    clock = attr.ib()
    _endpoints = attr.ib(default=attr.Factory(dict))

    def get_product_endpoints(self):
        """
        Return a dictionary of product endpoints registered with this class.
        """
        return dict(self._endpoints)

    def register_product(self, title, href):
        """
        If the product, identified by href, doesn't already exist in the
        registry, then register a product by creating a URL that one can GET
        and/or POST from/to, and providing a descriptive title for it in the
        product endpoint listing.

        If the product already appears in the registry, take no action.

        :param str title: This provides the human-readable description of the
            product to the user when they GET the list of endpoints.
        :param str href: This provides the unique component of an HTTP reference
            URL that a client can use to talk to this specific feed.
        """
        if not self.get_product_by_href(href):
            self._endpoints[href] = CloudFeedsProduct(title=title, href=href)

    def get_product_by_href(self, href):
        """
        If it exists, returns the product endpoint for the given href.
        If no such endpoint exists, return None.
        """
        return self._endpoints.get(href, None)


def render_product_dict(the_product):
    """
    Return Python dictionary suitable for JSON encoding in requests and
    responses to/from cloudfeeds.  This dictionary contains a single product
    endpoint descriptor.
    """
    return {
        "title": the_product.title,
        "collection": {
            "title": the_product.title,
            "href": the_product.href,
        }
    }


def render_product_endpoints_dict(the_products):
    """
    Return Python dictionary suitable for JSON encoding in requests and
    responses to/from cloudfeeds.  This dictionary contains the complete set
    of product endpoints.
    """
    return {
        "service": {
            "workspace": [
                render_product_dict(the_products[p]) for p in the_products
            ],
        },
    }



"""
Tests for :mod:`nova_api` and :mod:`nova_objects` for images.
"""

from twisted.trial.unittest import SynchronousTestCase

from mimic.test.helpers import json_request, request
from mimic.rest.nova_api import NovaApi, NovaControlApi
from mimic.test.fixtures import APIMockHelper
import random


class NovaAPIVirtualImageTests(SynchronousTestCase):

    """
    Tests for Virtual Images using the Nova Api plugin.
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`NovaApi` as the only plugin.
        """
        nova_api = NovaApi(["ORD"])
        self.helper = self.helper = APIMockHelper(
            self, [nova_api, NovaControlApi(nova_api=nova_api)]
        )
        self.root = self.helper.root
        self.uri = self.helper.uri

    def get_server_image(self, postfix):
        """
        Get images, assert response code is 200 and return response body.
        """
        (response, content) = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + postfix))
        self.assertEqual(200, response.code)
        return content

    def test_get_server_image_negative(self):
        """
        Test to verify :func:`get_image` when invalid image from the
        :obj: `mimic_presets` is provided.
        """
        get_server_image = request(self, self.root, "GET", self.uri +
                                   '/images/invalid_image_id')
        get_server_image_response = self.successResultOf(get_server_image)
        self.assertEqual(get_server_image_response.code, 404)

    def test_get_server_image_negative2(self):
        """
        Test to verify :func:`get_image` when invalid image from the
        :obj: `mimic_presets` is provided.
        """
        get_server_image = request(self, self.root, "GET", self.uri +
                                   '/images/any-image-ending-with-Z')
        get_server_image_response = self.successResultOf(get_server_image)
        self.assertEqual(get_server_image_response.code, 404)

    def test_get_image_that_does_not_exist(self):
        """
        Test to verify :func:`get_image` on ``GET /v2.0/<tenant_id>/images/<image_id>``
        """
        response, body = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/images/id_not_found'))

        self.assertEqual(response.code, 404)
        self.assertEqual(body, {
            "itemNotFound": {
                "message": "Image not found.",
                "code": 404
            }
        })

    def test_get_virtual_server_image(self):
        """
        Test to verify :func:`get_image` on ``GET /v2.0/<tenant_id>/images/<image_id>``
        """
        get_image_list_response_body = self.get_server_image('/images')
        image_list = get_image_list_response_body['images']
        choice = random.randint(0, len(image_list) - 1)
        image_id = image_list[choice]['id']
        get_image_response_body = self.get_server_image('/images/' + image_id)
        self.assertEqual(
            get_image_response_body['image']['id'], image_id)
        self.assertEqual(
            get_image_response_body['image']['status'], 'ACTIVE')

    def test_get_virtual_server_image_list(self):
        """
        Test to verify :func:`get_image_list` on ``GET /v2.0/<tenant_id>/images``
        """
        get_image_list_response_body = self.get_server_image('/images')
        image_list = get_image_list_response_body['images']
        self.assertEqual(len(image_list), 38)
        for each_image in image_list:
            self.assertEqual(sorted(each_image.keys()), sorted(['id', 'name', 'links']))

    def test_get_image_list_with_details(self):
        """
        Test to verify :func:`test_get_image_list_with_details` on
        ``GET /v2.0/<tenant_id>/images/details``
        """
        get_image_list_response_body = self.get_server_image('/images/detail')
        image_list = get_image_list_response_body['images']
        self.assertTrue(len(image_list) > 1)
        for each_image in image_list:
            self.assertEqual(
                sorted(each_image.keys()),
                sorted(['id', 'name', 'links', 'minRam', 'status',
                        'OS-EXT-IMG-SIZE:size', 'metadata', 'progress', 'created', 'updated',
                        'minDisk', 'com.rackspace__1__ui_default_show']))

    def test_get_image_list_with_details_is_consistent(self):
        """
        Test to verify ``GET /v2.0/<tenant_id>/images/details`` returns
        consistent data on mutiple list calls
        """
        get_image_list_response_body = self.get_server_image('/images/detail')
        get_image_list2_response_body = self.get_server_image('/images/detail')
        self.assertEqual(get_image_list_response_body, get_image_list2_response_body)

    def test_get_image_list_after_list_images(self):
        """
        Test to verify :func:`get_image_list` on ``GET /v2.0/<tenant_id>/images``
        """
        image_list = self.get_server_image('/images')['images']
        for each_image in image_list:
            get_server_image_response_body = self.get_server_image(
                '/images/' + each_image['id'])
            self.assertEqual(each_image['id'],
                             get_server_image_response_body['image']['id'])
            self.assertEqual(each_image['name'],
                             get_server_image_response_body['image']['name'])

    def test_no_OnMetal_images_in_ORD(self):
        """
        Test to verify OnMetal images are not in ORD region
        """
        get_image_list_response_body = self.get_server_image('/images/detail')
        image_list = get_image_list_response_body['images']
        for image in image_list:
            self.assertTrue('!onmetal' in image['metadata']['flavor_classes'])


class NovaAPIOnMetalImageTests(SynchronousTestCase):

    """
    Tests for Virtual Images using the Nova Api plugin.
    """

    def setUp(self):
        """
        Create a :obj:`MimicCore` with :obj:`NovaApi` as the only plugin.
        """
        nova_api = NovaApi(["IAD"])
        self.helper = self.helper = APIMockHelper(
            self, [nova_api, NovaControlApi(nova_api=nova_api)]
        )
        self.root = self.helper.root
        self.uri = self.helper.uri

    def get_server_image(self, postfix):
        """
        Get images, assert response code is 200 and return response body.
        """
        (response, content) = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + postfix))
        self.assertEqual(200, response.code)
        return content

    def test_get_OnMetal_server_image(self):
        """
        Test to verify :func:`get_image` on ``GET /v2.0/<tenant_id>/images/<image_id>``
        """
        get_image_list_response_body = self.get_server_image('/images/detail')
        image_list = get_image_list_response_body['images']
        for image in image_list:
            if image['metadata']['flavor_classes'] == 'onmetal':
                onmetal_id = image['id']
                break
        response, body = self.successResultOf(json_request(
            self, self.root, "GET", self.uri + '/images/' + onmetal_id))
        self.assertEqual(200, response.code)
        self.assertEqual(body['image']['id'], onmetal_id)
        self.assertEqual(body['image']['metadata']['flavor_classes'], 'onmetal')

    def test_OnMetal_image_list(self):
        """
        Test to verify :func:`images` on ``GET /v2.0/<tenant_id>/images``
        """
        get_image_list_response_body = self.get_server_image('/images')
        image_list = get_image_list_response_body['images']
        self.assertTrue(len(image_list), 52)
        for each_image in image_list:
            self.assertEqual(sorted(each_image.keys()), sorted(['id', 'links', 'name']))

    def test_OnMetal_image_list_details(self):
        """
        Test to verify :func:`get_image_list` on ``GET /v2.0/<tenant_id>/images``
        includes OnMetal images in IAD
        """
        get_image_list_response_body = self.get_server_image('/images/detail')
        image_list = get_image_list_response_body['images']
        self.assertTrue(len(image_list), 52)
        for each_image in image_list:
            self.assertEqual(sorted(each_image.keys()), sorted(['id', 'name', 'links', 'minRam',
                                                                'status', 'OS-EXT-IMG-SIZE:size',
                                                                'metadata', 'progress', 'created',
                                                                'updated', 'minDisk',
                                                                'com.rackspace__1__ui_default_show']))



from mimic.model import cloudfeeds
from twisted.trial.unittest import SynchronousTestCase
from testtools.matchers import (MatchesSetwise, MatchesDict, Equals)


class TestCloudFeeds(SynchronousTestCase):
    def setUp(self):
        self.cf = cloudfeeds.CloudFeeds(tenant_id='1234', clock=None)

    def test_creation(self):
        """
        A new CloudFeeds plugin should have no products when created.
        """
        self.assertEquals(len(self.cf.get_product_endpoints()), 0)

    def test_product_registration(self):
        """
        Registering a new product should create a new ATOM feed.
        """
        self.cf.register_product(title='The hoohaw product.', href='hoohaw')
        self.assertEquals(len(self.cf.get_product_endpoints()), 1)

    def test_product_reregistration(self):
        """
        Re-registering a new product should do nothing.
        """
        self.cf.register_product(title='The hoohaw product', href='hoohaw')
        self.cf.register_product(title='The OTHER hoohaw product', href='hoohaw')
        self.assertEquals(len(self.cf.get_product_endpoints()), 1)
        p = self.cf.get_product_by_href('hoohaw')
        self.assertEquals(p.title, 'The hoohaw product')

    def test_get_products(self):
        """
        Requesting a list of product endpoints should return a title and an href
        for each endpoint.
        """
        self.cf.register_product(title='The hoohaw product', href='hoohaw')
        self.cf.register_product(title='The goober product', href='goober')
        products = self.cf.get_product_endpoints()
        self.assertEquals('hoohaw' in products, True)
        self.assertEquals(products['hoohaw'].title, 'The hoohaw product')
        self.assertEquals(products['goober'].title, 'The goober product')
        self.assertEquals(products['hoohaw'].href, 'hoohaw')
        self.assertEquals(products['goober'].href, 'goober')


class TestCloudFeedsProduct(SynchronousTestCase):
    def test_creation(self):
        """
        A new product queue should be empty.
        """
        cfp = cloudfeeds.CloudFeedsProduct(title='title', href='href')
        self.assertEquals(len(cfp.events), 0)

    def test_post(self):
        """
        Posting a new event to a queue should tack said event onto the end
        of said queue.
        """
        cfp = cloudfeeds.CloudFeedsProduct(title='title', href='href')
        cfp.post("TROLOLOLOLOL!!!")
        cfp.post("This is a totally fake event-like thing.")
        self.assertEquals(
            cfp.events,
            ["TROLOLOLOLOL!!!", "This is a totally fake event-like thing."]
        )


class TestSerialization(SynchronousTestCase):
    def test_json_description(self):
        """
        When listing product endpoints, we expect our JSON to look a certain
        way.  The easiest way to do that is to acquire the corresponding
        dict, then pass it through json.dump with your preferred formatting
        settings.
        """
        cfp = cloudfeeds.CloudFeedsProduct(title='title', href='href')
        d = cloudfeeds.render_product_dict(cfp)
        productDescription = MatchesDict({
            "title": Equals("title"),
            "collection": MatchesDict({
                "href": Equals("href"),
                "title": Equals("title"),
            }),
        })
        self.assertEquals(productDescription.match(d), None)

    def test_json_product_list(self):
        """
        When listing product endpoints, the resulting JSON should contain a
        service object, and a workspace object, and within that, an array of
        product descriptors.
        """
        cf = cloudfeeds.CloudFeeds(tenant_id='1234', clock=None)
        cf.register_product(title="The hoohaw product", href="hoohaw")
        cf.register_product(title="The goober product", href="goober")
        listing = MatchesDict({
            "service": MatchesDict({
                "workspace": MatchesSetwise(
                    MatchesDict({
                        "collection": MatchesDict({
                            "href": Equals("hoohaw"),
                            "title": Equals("The hoohaw product"),
                        }),
                        "title": Equals("The hoohaw product"),
                    }),
                    MatchesDict({
                        "collection": MatchesDict({
                            "href": Equals("goober"),
                            "title": Equals("The goober product"),
                        }),
                        "title": Equals("The goober product"),
                    }),
                ),
            })
        })
        self.assertEquals(
            listing.match(cloudfeeds.render_product_endpoints_dict(
                cf.get_product_endpoints()
            )),
            None
        )



"""
Canned responses for MAAS monitoring zones.
"""


def monitoring_zones():
    """
    Canned response for the /monitoring_zones call.
    """
    return [{
        "id": "mzdfw",
        "label": "Dallas Fort Worth (DFW)",
        "country_code": "US",
        "source_ips": [
            "dead:beef:cafe:face::/64",
            "1.2.3.128/26"
        ]
    }, {
        "id": "mzhkg",
        "label": "Hong Kong (HKG)",
        "country_code": "HK",
        "source_ips": [
            "4.5.6.64/26",
            "1337:f005:ba11:d00d:0:0:0:0/64"
        ]
    }, {
        "id": "mziad",
        "label": "Northern Virginia (IAD)",
        "country_code": "US",
        "source_ips": [
            "5ca1:ab1e:7e1e:ca57::/64",
            "7.8.9.192/26"
        ]
    }, {
        "id": "mzlon",
        "label": "London (LON)",
        "country_code": "GB",
        "source_ips": [
            "1ceb:00da:8bad:food::/64",
            "11.12.13.0/26"
        ]
    }, {
        "id": "mzord",
        "label": "Chicago (ORD)",
        "country_code": "US",
        "source_ips": [
            "ba5e:ba11:1111:2222::/64",
            "14.15.16.0/26"
        ]
    }, {
        "id": "mzsyd",
        "label": "Sydney (SYD)",
        "country_code": "AU",
        "source_ips": [
            "17.18.19.0/26",
            "1111:2222:3333:4444::/64"
        ]
    }]



"""
Plugin for Rackspace cloud feeds mock.
"""
from mimic.rest.cloudfeeds import CloudFeedsApi, CloudFeedsControlApi

cloudfeeds = CloudFeedsApi()
cloudfeeds_control = CloudFeedsControlApi(cf_api=cloudfeeds)



"""
Mimic twisted application plugins.
"""
from twisted.application.service import ServiceMaker


mimicService = ServiceMaker(
    "mimic Service/",
    "mimic.tap",
    "Mocks for Rackspace APIs",
    "mimic")



"""
Model objects for the Glance mimic.
"""
from json import dumps, loads
from characteristic import attributes, Attribute
from uuid import uuid4

random_image_list = [
    {"id": str(uuid4()), "name": "OnMetal - CentOS 6", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - CentOS 7", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - CoreOS (Alpha)", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - CoreOS (Beta)", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - Debian 7 (Wheezy)", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - Debian 8 (Jessie)", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - Fedora 21", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - Fedora 22", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - Ubuntu 14.04 LTS (Trusty Tahr)", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - CoreOS (Stable)", "distro": "linux"},
    {"id": str(uuid4()), "name": "OnMetal - Ubuntu 12.04 LTS (Precise Pangolin)",
     "distro": "linux"},
    {"id": str(uuid4()), "name": "Ubuntu 14.04 LTS (Trusty Tahr)", "distro": "linux"},
    {"id": str(uuid4()), "name": "Ubuntu 15.04 (Vivid Vervet)", "distro": "linux"},
    {"id": str(uuid4()), "name": "Windows Server 2012 R2", "distro": "windows"}
]


@attributes(["image_id", "name", "distro",
             Attribute("tenant_id", default_value=None),
             Attribute("status", default_value='active')])
class Image(object):
    """
    A Image object
    """

    static_server_image_defaults = {
        "minRam": 256,
        "minDisk": 00,
    }

    static_glance_defaults = {
        "flavor_classes": "*",
        "min_ram": 256,
        "min_disk": 00,
        "container_format": None,
        "owner": "00000",
        "size": 10000,
        "tags": [],
        "visibility": "public",
        "checksum": "0000",
        "protected": False,
        "disk_format": None,
        "ssh_user": "mimic",
        "schema": "/v2/schemas/image",
        "auto_disk_config": "disabled",
        "virtual_size": None,
        "visibility": "public"
    }

    static_metadata = {
        "com.rackspace__1__build_rackconnect": "1",
        "com.rackspace__1__options": "0",
        "com.rackspace__1__release_id": "000",
        "com.rackspace__1__build_core": "1",
        "image_type": "base",
        "org.openstack__1__os_version": "0.1",
        "com.rackspace__1__platform_target": "MimicCloud",
        "com.rackspace__1__build_managed": "1",
        "org.openstack__1__architecture": "x64",
        "com.rackspace__1__visible_core": "1",
        "com.rackspace__1__release_build_date": "1972-01-01_15-59-11",
        "com.rackspace__1__visible_rackconnect": "1",
        "com.rackspace__1__release_version": "1",
        "com.rackspace__1__visible_managed": "1",
        "cache_in_nova": "True",
        "com.rackspace__1__build_config_options": "mimic",
        "auto_disk_config": "True",
        "com.rackspace__1__source": "kickstart",
        "com.rackspace__1__ui_default_show": "True"
    }

    def links_json(self, absolutize_url):
        """
        Create a JSON-serializable data structure describing the links to this
        image.
        """
        return [
            {
                "href": absolutize_url("v2/{0}/images/{1}"
                                       .format(self.tenant_id, self.image_id)),
                "rel": "self"
            },
            {
                "href": absolutize_url("{0}/images/{1}"
                                       .format(self.tenant_id, self.image_id)),
                "rel": "bookmark"
            },
            {
                "href": absolutize_url("/images/{0}".format(self.image_id)),
                "rel": "alternate",
                "type": "application/vnd.openstack.image"
            }
        ]

    def get_server_image_details_json(self, absolutize_url):
        """
        JSON-serializable object representation of this image, as
        returned by either a GET on this individual image through the
        servers api.
        """
        template = self.static_server_image_defaults.copy()
        template.update({
            "id": self.image_id,
            "name": self.name,
            "status": self.status.upper(),
            "links": self.links_json(absolutize_url),
            "progress": 100,
            "OS-DCF:diskConfig": "AUTO",
            "OS-EXT-IMG-SIZE:size": 100000,
            "metadata": self.static_metadata,
            "created": "1972-01-01_15-59-11",
            "updated": "1972-01-01_15-59-11"
        })
        if self.distro != "windows":
            template["metadata"]["os_distro"] = self.distro
        return template

    def brief_json(self, absolutize_url):
        """
        Brief JSON-serializable version of this image.
        """
        return {
            "name": self.name,
            "id": self.image_id,
            "links": self.links_json(absolutize_url)
        }

    def get_glance_admin_image_json(self):
        """
        JSON-serializable object representation of this image, as
        returned by either a GET on this individual image or a member in the
        list returned by the list-details request.
        """
        template = self.static_glance_defaults.copy()
        template.update(self.static_metadata)
        template.update({
            "id": self.image_id,
            "name": self.name,
            "status": self.status,
            "created_at": "1972-01-01_15-59-11",
            "updated_at": "1972-01-01_15-59-11",
            "file": "/v2/images/{0}/file".format(self.image_id),
            "self": "/v2/images/" + self.image_id,
            "org.openstack__1__os_distro": "mimic." + self.distro,
            "os_type": self.distro,
            "vm_mode": "onmetal"
        })
        if self.distro != "windows":
            template.update({
                "os_distro": self.distro
            })
        if "OnMetal" in self.name:
            template.update({
                "vm_mode": "metal",
                "flavor_classes": "onmetal"
            })
        return template


@attributes([Attribute("glance_admin_image_store", default_factory=list)])
class GlanceAdminImageStore(object):
    """
    A collection of :obj:`Image`.
    """
    def image_by_id(self, image_id):
        """
        Retrieve a :obj:`Image` object by its ID.
        """
        for image in self.glance_admin_image_store:
            if image.image_id == image_id:
                return image

    def add_to_glance_admin_image_store(self, **attributes):
        """
        Create a new Image object and add it to the
        :obj: `glance_admin_image_store`
        """
        image = Image(**attributes)
        self.glance_admin_image_store.append(image)
        return image

    def list_images(self):
        """
        List all the images for the Glance Admin API.
        """
        if not self.glance_admin_image_store:
            for each_image in random_image_list:
                self.add_to_glance_admin_image_store(
                    image_id=each_image['id'],
                    name=each_image['name'],
                    distro=each_image['distro'])
        return {"images": [image.get_glance_admin_image_json()
                           for image in self.glance_admin_image_store]}

    def get_image(self, http_request, image_id):
        """
        get image with image_id for the Glance Admin API.
        """
        image = self.image_by_id(image_id)
        if image:
            return image.get_glance_admin_image_json()
        http_request.setResponseCode(404)
        return b''

    def create_image(self, http_create_request):
        """
        Creates a new image with the given request json and returns the image.

        Note: This is more like a control plane API as I dint find seem
        to find documentation for add image under the Glance admin API.
        """
        content = loads(http_create_request.content.read())
        try:
            image_id = str(uuid4())
            new_image = self.add_to_glance_admin_image_store(
                image_id=image_id,
                name=content.get('name'),
                distro=content.get('distro'))
            http_create_request.setResponseCode(201)
            return new_image.get_glance_admin_image_json()
        except Exception as e:
            http_create_request.setResponseCode(400)
            return dumps({"Error": str(e)})

    def delete_image(self, http_request, image_id):
        """
        Deletes the image and returns 204.
        If image does not exit, returns 404.
        Docs: http://bit.ly/1Obujvd
        """
        image = self.image_by_id(image_id)
        if image:
            self.glance_admin_image_store.remove(image)
            http_request.setResponseCode(204)
            return b''
        http_request.setResponseCode(404)
        return b''




import six
from datetime import datetime
import re

from twisted.trial.unittest import SynchronousTestCase

from twisted.internet.task import Clock

from mimic.session import NonMatchingTenantError, SessionStore


class SessionCreationTests(SynchronousTestCase):
    """
    Tests for :class:`SessionStore`.
    """

    def test_username_password_new(self):
        """
        SessionStore.session_for_username_password creates a new session (if no
        such session exists for the given username).
        """
        clock = Clock()
        sessions = SessionStore(clock)
        clock.advance(4321)
        session = sessions.session_for_username_password("example_user",
                                                         "password")
        self.assertEqual(session.username, "example_user")
        self.assertEqual(session.expires,
                         datetime.utcfromtimestamp(4321 + 86400))
        self.assertIsInstance(session.tenant_id, six.text_type)
        self.assertIsInstance(session.token, six.text_type)
        self.assertNotEqual(session.username, session.token)
        self.assertNotEqual(session.token, session.tenant_id)

    def test_username_password_wrong_tenant(self):
        """
        Tenant ID is validated in
        :func:`SessionStore.session_for_username_password`.

        If called with the token of an existing session but the wrong tenant,
        raises :class:`NonMatchingTenantError`.
        """
        """
        SessionStore.session_for_username_password, if called with the
        username of an existing session but the wrong tenant, raises
        :class:`NonMatchingTenantError`.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        clock.advance(4321)
        sessions.session_for_username_password("example_user",
                                               "password",
                                               "tenant_orig")
        self.assertRaises(
            NonMatchingTenantError,
            sessions.session_for_username_password,
            "example_user", "password", "tenant_new")

    def test_different_username_different_token(self):
        """
        Sessions are distinct if they are requested with distinct usernames.
        """
        sessions = SessionStore(Clock())
        a = sessions.session_for_username_password("a", "ignored")
        b = sessions.session_for_username_password("b", "ignored")
        self.assertNotEqual(a.token, b.token)

    def test_by_username_after_token(self):
        """
        SessionStore.session_for_username_password should retrieve the same
        session that was created by SessionStore.session_for_token.  Similarly
        for the API key.
        """
        sessions = SessionStore(Clock())
        a = sessions.session_for_token("testtoken")
        b = sessions.session_for_username_password(a.username, "testpswd")
        c = sessions.session_for_api_key(a.username, "testapikey")
        self.assertIdentical(a, b)
        self.assertIdentical(a, c)

    def test_session_for_apikey_after_username_wrong_tenant(self):
        """
        Tenant ID is validated in :func:`SessionStore.session_for_api_key`.

        If called with the token of an existing session but the wrong tenant,
        raises :class:`NonMatchingTenantError`.
        """
        sessions = SessionStore(Clock())
        a = sessions.session_for_username_password("username", "testpswd")
        self.assertRaises(
            NonMatchingTenantError,
            sessions.session_for_api_key,
            a.username, "testapikey", a.tenant_id + "wrong")

    def test_by_token_after_username(self):
        """
        Session retrieved by all the ``session_for_*`` methods are identical.

        :func:`SessionStore.session_for_token` should retrieve the same
        session that was created by
        :func:`SessionStore.session_for_username_password`.
        """
        sessions = SessionStore(Clock())
        a = sessions.session_for_username_password("username",
                                                   "testpswd")
        b = sessions.session_for_token(a.token)
        self.assertIdentical(a, b)
        c = sessions.session_for_api_key("apiuser", "testkey")
        d = sessions.session_for_token(c.token)
        self.assertIdentical(c, d)

    def test_by_token_after_username_wrong_tenant(self):
        """
        Tenant ID is validated in :func:`SessionStore.session_for_token`.

        If called with the token of an existing session but the wrong tenant,
        raises :class:`NonMatchingTenantError`.
        """
        sessions = SessionStore(Clock())
        a = sessions.session_for_username_password("username",
                                                   "testpswd")
        self.assertRaises(
            NonMatchingTenantError,
            sessions.session_for_token,
            a.token, a.tenant_id + 'wrong')

    def test_impersonation(self):
        """
        SessionStore.session_for_impersonation will return a session that can
        be retrieved by impersonated token_id or username.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        A_LITTLE = 1234
        clock.advance(A_LITTLE)
        A_LOT = 65432
        a = sessions.session_for_impersonation("pretender", A_LOT)
        a_prime = sessions.session_for_impersonation("pretender", A_LOT)
        self.assertIdentical(a, a_prime)
        b = sessions.session_for_token(a.token)
        self.assertEqual(
            a.expires, datetime.utcfromtimestamp(A_LITTLE + A_LOT))
        self.assertIdentical(a, b)
        c = sessions.session_for_username_password("pretender",
                                                   "not a password")
        self.assertIdentical(a, c)
        self.assertEqual(a.username, c.username)
        self.assertEqual(a.tenant_id, c.tenant_id)

        # Right now all data_for_api cares about is hashability; this may need
        # to change if it comes to rely upon its argument actually being an API
        # mock.
        same_api = 'not_an_api'

        username_data = c.data_for_api(same_api, list)
        token_data = b.data_for_api(same_api, list)
        impersonation_data = a.data_for_api(same_api, list)

        self.assertIs(username_data, impersonation_data)
        self.assertIs(token_data, impersonation_data)

    def test_session_for_tenant_id(self):
        """
        SessionStore.session_for_tenant_id will return a session that can be
        retrieved by tenant_id.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        session = sessions.session_for_username_password("someuser",
                                                         "testpass")
        session2 = sessions.session_for_tenant_id(session.tenant_id)
        self.assertIdentical(session, session2)

    def test_generate_username_from_tenant_id(self):
        """
        SessionStore.session_for_tenant_id will create a new session with a
        synthetic username if no such tenant ID yet exists.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        session = sessions.session_for_tenant_id("user_specified_tenant")
        session2 = sessions.session_for_username_password(session.username,
                                                          "testpass")
        self.assertIdentical(session, session2)

    def test_session_for_tenant_id_with_custom_tenant(self):
        """
        SessionStore.session_for_tenant_id will return a session that can be
        retrieved by tenant_id.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        session = sessions.session_for_username_password(
            "someuser", "testpass", "sometenant"
        )
        session2 = sessions.session_for_tenant_id("sometenant")
        self.assertIdentical(session, session2)

    def test_sessions_created_all_have_integer_tenant_ids(self):
        """
        Sessions created by
        :class:`SessionStore.session_for_username_password`,
        :class:`SessionStore.session_for_impersonation`,
        :class:`SessionStore.session_for_api_key`, and
        :class:`SessionStore.session_for_token`, when not passed a specific
        tenant ID, all generate integer-style tenant IDs.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        sessions = [
            sessions.session_for_username_password("someuser1", "testpass"),
            sessions.session_for_impersonation("someuser2", 12),
            sessions.session_for_api_key("someuser3", "someapikey"),
            sessions.session_for_token("sometoken"),
        ]
        integer = re.compile('^\d+$')
        for session in sessions:
            self.assertIsNot(integer.match(session.tenant_id), None,
                             "{0} is not an integer.".format(
                                 session.tenant_id))
            self.assertTrue(int(session.tenant_id) < 1e15)

    def test_sessions_created_honor_specified_tenant_id(self):
        """
        Sessions created by
        :class:`SessionStore.session_for_username_password`,
        :class:`SessionStore.session_for_api_key`, and
        :class:`SessionStore.session_for_token`,
        :class:`SessionStore.session_for_tenant_id` all honor the passed-in
        tenant ID.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        sessions = [
            sessions.session_for_username_password("user1", "pass",
                                                   "tenant1"),
            sessions.session_for_api_key("user2", "apikey",
                                         tenant_id="tenant2"),
            sessions.session_for_token("token", tenant_id="tenant3"),
            sessions.session_for_tenant_id("tenant4")
        ]
        for i, session in enumerate(sessions):
            self.assertEqual("tenant{0}".format(i + 1), session.tenant_id)

    def test_token_after_api_key_specifying_tenant(self):
        """
        Sessions created by
        :class:`SessionStore.session_for_api_key` and specifying
        the tenant ID should be returned on requests to
        :class:`SessionStore.session_for_token` that also specify
        the same tenant ID.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        session_by_api_key = sessions.session_for_api_key(
            "user1", "f005ba11", tenant_id="559638")
        session_by_token = sessions.session_for_token(
            "token", tenant_id="559638")
        self.assertIs(session_by_api_key, session_by_token)

    def test_username_password_after_token_specifying_tenant(self):
        """
        Sessions created by
        :class:`SessionStore.session_for_token` and specifying
        the tenant ID should be returned on requests to
        :class:`SessionStore.session_for_username_password` that
        also specify the same tenant ID.
        """
        clock = Clock()
        sessions = SessionStore(clock)
        session_by_token = sessions.session_for_token("token", tenant_id="tenant1337")
        session_by_username_password = sessions.session_for_username_password(
            "user1", "pass", "tenant1337")
        self.assertIs(session_by_token, session_by_username_password)



# -*- test-case-name: mimic.test.test_nova -*-
"""
Defines create, delete, get, list servers and get images and flavors.
"""

from uuid import uuid4
import json

from characteristic import attributes
from six import text_type

from zope.interface import implementer

from twisted.python.urlpath import URLPath

from twisted.plugin import IPlugin
from twisted.web.http import CREATED, BAD_REQUEST

from mimic.canned_responses.nova import get_limit
from mimic.model.keypair_objects import GlobalKeyPairCollections, KeyPair
from mimic.rest.mimicapp import MimicApp
from mimic.catalog import Entry
from mimic.catalog import Endpoint
from mimic.imimic import IAPIMock
from mimic.model.behaviors import make_behavior_api
from mimic.model.nova_objects import (
    BadRequestError, GlobalServerCollections, LimitError, Server,
    bad_request, forbidden, not_found, server_creation)
from mimic.model.flavor_collections import GlobalFlavorCollection
from mimic.model.image_collections import GlobalImageCollection


@implementer(IAPIMock, IPlugin)
class NovaApi(object):

    """
    Rest endpoints for mocked Nova Api.
    """

    def __init__(self, regions=["ORD", "IAD", "DFW"]):
        """
        Create a NovaApi with an empty region cache, no servers or tenants yet.
        """
        self._regions = regions

    def catalog_entries(self, tenant_id):
        """
        List catalog entries for the Nova API.
        """
        return [
            Entry(
                tenant_id, "compute", "cloudServersOpenStack",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()),
                             prefix="v2")
                    for region in self._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        return (NovaRegion(self, uri_prefix, session_store, region)
                .app.resource())

    def _get_session(self, session_store, tenant_id):
        """
        Retrieve or create a new Nova session from a given tenant identifier
        and :obj:`SessionStore`.

        For use with ``data_for_api``.

        Temporary hack; see this issue
        https://github.com/rackerlabs/mimic/issues/158
        """
        return (
            session_store.session_for_tenant_id(tenant_id)
            .data_for_api(self, lambda: GlobalServerCollections(
                tenant_id=tenant_id,
                clock=session_store.clock
            ))
        )


@implementer(IAPIMock, IPlugin)
@attributes(["nova_api"])
class NovaControlApi(object):

    """
    Rest endpoints for the Nova Control Api.
    """

    def catalog_entries(self, tenant_id):
        """
        List catalog entries for the Nova API.
        """
        return [
            Entry(
                tenant_id, "compute", "cloudServersBehavior",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()),
                             prefix="v2")
                    for region in self.nova_api._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        return (NovaControlApiRegion(api_mock=self, uri_prefix=uri_prefix,
                                     session_store=session_store, region=region)
                .app.resource())


NovaControlApiRegionBehaviors = make_behavior_api(
    {'creation': server_creation})
"""
Handlers for CRUD operations on create server behaviors.

:ivar registry_collection: an instance of
    :class:`mimic.model.behaviors.BehaviorRegistryColelction`
"""


@attributes(["api_mock", "uri_prefix", "session_store", "region"])
class NovaControlApiRegion(object):
    """
    Klein resources for the Nova Control plane API
    """
    app = MimicApp()

    @app.route('/v2/<string:tenant_id>/behaviors', branch=True)
    def handle_behaviors(self, request, tenant_id):
        """
        Return the behavior CRUD resource.
        """
        region_collection = self._collection_from_tenant(tenant_id)
        behavior_handler = NovaControlApiRegionBehaviors(
            region_collection.behavior_registry_collection)
        return behavior_handler.app.resource()

    @app.route("/v2/<string:tenant_id>/attributes/", methods=['POST'])
    def change_attributes(self, request, tenant_id):
        """
        Modify the specified attributes on existing servers.

        The request looks like this::

            {
                "status": {
                    "test_server_id_8482197407": "ERROR",
                    "test_server_id_0743289146": "BUILD"
                }
            }

        As more attributes are added, they should be additional top-level keys
        where "status" goes in this request.
        """
        region_collection = self._collection_from_tenant(tenant_id)
        attributes_description = json.loads(request.content.read())
        statuses_description = attributes_description["status"]
        servers = [region_collection.server_by_id(server_id)
                   for server_id in statuses_description]
        if None in servers:
            request.setResponseCode(BAD_REQUEST)
            return b''
        for server in servers:
            server.update_status(statuses_description[server.server_id])
        request.setResponseCode(CREATED)
        return b''

    def _collection_from_tenant(self, tenant_id):
        """
        Retrieve the server collection for this region for the given tenant.
        """
        return (self.api_mock.nova_api._get_session(self.session_store, tenant_id)
                .collection_for_region(self.region))


class NovaRegion(object):

    """
    Klein routes for the API within a Cloud Servers region.

    :ivar dict _tenant_cache: a mapping of tenant_id (bytes) to a "server
        cache" (:obj:`S_Cache`), which itself maps server_id to a
        JSON-serializable data structure of the 'server' key of GET responses.
    """

    def __init__(self, api_mock, uri_prefix, session_store, name):
        """
        Create a nova region with a given URI prefix (used for generating URIs
        to servers).
        """
        self.uri_prefix = uri_prefix
        self._api_mock = api_mock
        self._session_store = session_store
        self._name = name

    def url(self, suffix):
        """
        Generate a URL to an object within the Nova URL hierarchy, given the
        part of the URL that comes after.
        """
        return str(URLPath.fromString(self.uri_prefix).child(suffix))

    def _region_collection_for_tenant(self, tenant_id):
        """
        Get the given server-cache object for the given tenant, creating one if
        there isn't one.
        """
        return (self._api_mock._get_session(self._session_store, tenant_id)
                .collection_for_region(self._name))

    def _image_collection_for_tenant(self, tenant_id):
        tenant_session = self._session_store.session_for_tenant_id(tenant_id)
        image_global_collection = tenant_session.data_for_api(
            "image_collection",
            lambda: GlobalImageCollection(tenant_id=tenant_id,
                                          clock=self._session_store.clock))
        image_region_collection = image_global_collection.collection_for_region(
            self._name)
        return image_region_collection

    def _keypair_collection_for_tenant(self, tenant_id):
        """
        Returns the keypairs for a region
        """
        tenant_session = self._session_store.session_for_tenant_id(tenant_id)
        kp_global_collection = tenant_session.data_for_api(
            "keypair_collection",
            lambda: GlobalKeyPairCollections(tenant_id=tenant_id,
                                             clock=self._session_store.clock))
        kp_region_collection = kp_global_collection.collection_for_region(
            self._name)
        return kp_region_collection

    app = MimicApp()

    @app.route('/v2/<string:tenant_id>/servers', methods=['POST'])
    def create_server(self, request, tenant_id):
        """
        Returns a generic create server response, with status 'ACTIVE'.
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            return json.dumps(
                bad_request("Invalid JSON request body", request))

        try:
            creation = (self._region_collection_for_tenant(tenant_id)
                        .request_creation(request, content, self.url))
        except BadRequestError as e:
            return json.dumps(bad_request(e.nova_message, request))
        except LimitError as e:
            return json.dumps(forbidden(e.nova_message, request))

        return creation

    @app.route('/v2/<string:tenant_id>/servers/<string:server_id>', methods=['GET'])
    def get_server(self, request, tenant_id, server_id):
        """
        Returns a generic get server response, with status 'ACTIVE'
        """
        return (
            self._region_collection_for_tenant(tenant_id)
            .request_read(request, server_id, self.url)
        )

    @app.route('/v2/<string:tenant_id>/servers', methods=['GET'])
    def list_servers(self, request, tenant_id):
        """
        Returns list of servers that were created by the mocks, with the given
        name.
        """
        return (
            self._region_collection_for_tenant(tenant_id)
            .request_list(
                request, include_details=False, absolutize_url=self.url,
                name=request.args.get('name', [u""])[0],
                limit=request.args.get('limit', [None])[0],
                marker=request.args.get('marker', [None])[0]
            )
        )

    @app.route('/v2/<string:tenant_id>/servers/detail', methods=['GET'])
    def list_servers_with_details(self, request, tenant_id):
        """
        Returns list of servers that were created by the mocks, with details
        such as the metadata.
        """
        return (
            self._region_collection_for_tenant(tenant_id)
            .request_list(
                request, include_details=True, absolutize_url=self.url,
                name=request.args.get('name', [u""])[0],
                limit=request.args.get('limit', [None])[0],
                marker=request.args.get('marker', [None])[0],
                changes_since=request.args.get('changes-since', [None])[0]
            )
        )

    @app.route('/v2/<string:tenant_id>/servers/<string:server_id>',
               methods=['DELETE'])
    def delete_server(self, request, tenant_id, server_id):
        """
        Returns a 204 response code, for any server id'
        """
        return (
            self._region_collection_for_tenant(tenant_id)
            .request_delete(request, server_id)
        )

    @app.route('/v2/<string:tenant_id>/images/<string:image_id>', methods=['GET'])
    def get_image(self, request, tenant_id, image_id):
        """
        Returns a get image response, for any given imageid
        """
        return(self._image_collection_for_tenant(tenant_id)
               .get_image(request, image_id, absolutize_url=self.url))

    @app.route('/v2/<string:tenant_id>/images/detail', methods=['GET'])
    def get_server_image_list_with_details(self, request, tenant_id):
        """
        Returns a image list.
        """
        return (self._image_collection_for_tenant(tenant_id)
                .list_images(include_details=True, absolutize_url=self.url))

    @app.route('/v2/<string:tenant_id>/images', methods=['GET'])
    def get_server_image_list(self, request, tenant_id):
        """
        Returns a image list.
        """
        return(self._image_collection_for_tenant(tenant_id)
               .list_images(include_details=False, absolutize_url=self.url))

    @app.route('/v2/<string:tenant_id>/flavors/<string:flavor_id>', methods=['GET'])
    def get_flavor_details(self, request, tenant_id, flavor_id):
        """
        Returns a get flavor response, for any given flavorid
        """
        flavor_collection = GlobalFlavorCollection(tenant_id=tenant_id,
                                                   clock=self._session_store.clock)
        return (flavor_collection.collection_for_region(region_name=self._name)
                .get_flavor(request, flavor_id, absolutize_url=self.url))

    @app.route('/v2/<string:tenant_id>/flavors', methods=['GET'])
    def get_flavor_list(self, request, tenant_id):
        """
        Returns a list of flavor with the response code 200.
        docs: http://bit.ly/1eXTSDC
        TO DO: The length of flavor list can be set using the control plane.
               Also be able to set different flavor types in the future.
        """
        flavor_collection = GlobalFlavorCollection(tenant_id=tenant_id,
                                                   clock=self._session_store.clock)
        return (flavor_collection.collection_for_region(region_name=self._name)
                .list_flavors(include_details=False, absolutize_url=self.url))

    @app.route('/v2/<string:tenant_id>/flavors/detail', methods=['GET'])
    def get_flavor_list_with_details(self, request, tenant_id):
        """
        Returns a list of flavor details with the response code 200.
        TO DO: The length of flavor list can be set using the control plane.
               Also be able to set different flavor types in the future.
        """
        flavor_collection = GlobalFlavorCollection(tenant_id=tenant_id,
                                                   clock=self._session_store.clock)
        return (flavor_collection.collection_for_region(region_name=self._name)
                .list_flavors(include_details=True, absolutize_url=self.url))

    @app.route('/v2/<string:tenant_id>/limits', methods=['GET'])
    def get_limit(self, request, tenant_id):
        """
        Returns a get flavor response, for any given flavorid
        """
        request.setResponseCode(200)
        return json.dumps(get_limit())

    @app.route('/v2/<string:tenant_id>/servers/<string:server_id>/ips', methods=['GET'])
    def get_ips(self, request, tenant_id, server_id):
        """
        Returns the IP addresses for the specified server.
        """
        return (
            self._region_collection_for_tenant(tenant_id).request_ips(
                request, server_id
            )
        )

    @app.route('/v2/<string:tenant_id>/servers/<string:server_id>/metadata',
               branch=True)
    def handle_server_metadata(self, request, tenant_id, server_id):
        """
        Handle metadata requests associated with a particular server.  Server
        not found error message verified as of 2015-04-23 against Rackspace
        Nova.
        """
        server = (self._region_collection_for_tenant(tenant_id)
                  .server_by_id(server_id))
        if server is None:
            return json.dumps(not_found('Server does not exist', request))

        return ServerMetadata(server).app.resource()

    @app.route('/v2/<string:tenant_id>/servers/<string:server_id>/action', methods=['POST'])
    def perform_action(self, request, tenant_id, server_id):
        """
        Perform the requested action on the server
        """
        return self._region_collection_for_tenant(tenant_id).request_action(request, server_id, self.url)

    @app.route("/v2/<string:tenant_id>/os-keypairs", methods=['GET'])
    def get_key_pairs(self, request, tenant_id):
        """
        Returns current key pairs.
        http://docs.rackspace.com/servers/api/v2/cs-devguide/content/ListKeyPairs.html
        """
        return self._keypair_collection_for_tenant(tenant_id).json_list()

    @app.route("/v2/<string:tenant_id>/os-keypairs", methods=['POST'])
    def create_key_pair(self, request, tenant_id):
        """
        Returns a newly created key pair with the specified name.
        http://docs.rackspace.com/servers/api/v2/cs-devguide/content/UploadKeyPair.html
        """
        try:
            content = json.loads(request.content.read())
            keypair = content["keypair"]
            keypair_from_request = KeyPair(
                name=keypair["name"], public_key=keypair["public_key"])
        except (ValueError or KeyError):
            request.setResponseCode(400)
            return json.dumps(bad_request("Malformed request body", request))

        keypair_response = self._keypair_collection_for_tenant(
            tenant_id).create_keypair(keypair=keypair_from_request)
        return json.dumps(keypair_response)

    @app.route("/v2/<string:tenant_id>/os-keypairs/<string:keypairname>", methods=['DELETE'])
    def delete_key_pair(self, request, tenant_id, keypairname):
        """
        Removes a key by its name.
        http://docs.rackspace.com/servers/api/v2/cs-devguide/content/DeleteKeyPair.html
        """
        try:
            self._keypair_collection_for_tenant(
                tenant_id).remove_keypair(keypairname)
        except ValueError:
            request.setResponseCode(404)
            return json.dumps("KeyPair not found: " + keypairname)

        request.setResponseCode(202)


class ServerMetadata(object):

    """
    Klein routes for a particular server's metadata.
    """

    def __init__(self, server):
        """
        Handle requests having to deal a server's metadata.  No URIs
        are generated or used.
        """
        self._server = server

    app = MimicApp()

    @app.route('/', methods=['GET'])
    def list_metadata(self, request):
        """
        List all metadata associated with a server.
        """
        return json.dumps({'metadata': self._server.metadata})

    @app.route('/', methods=['PUT'])
    def set_metadata(self, request):
        """
        Set the metadata for the specified server - this replaces whatever
        metadata was there.  The resulting metadata is not a union of the
        previous metadata and the new metadata - it is *just* the new
        metadata.

        The body must look like:

        ``{"metadata": {...}}``

        although

        ``{"metadata": {...}, "other": "garbage", "keys": "included"}``

        is ok too.

        All the response messages and codes have been verified as of
        2015-04-23 against Rackspace Nova.
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            return json.dumps(bad_request("Malformed request body", request))

        # more than one key is ok, non-"meta" keys are just ignored
        if 'metadata' not in content:
            return json.dumps(bad_request("Malformed request body", request))

        # When setting metadata, None is special for some reason
        if content['metadata'] is None:
            return json.dumps(bad_request(
                "Malformed request body. metadata must be object", request))

        try:
            Server.validate_metadata(content['metadata'])
        except BadRequestError as e:
            return json.dumps(bad_request(e.nova_message, request))
        except LimitError as e:
            return json.dumps(forbidden(e.nova_message, request))

        self._server.set_metadata(content['metadata'])
        return json.dumps({'metadata': content['metadata']})

    @app.route('/<key>', methods=['PUT'])
    def set_metadata_item(self, request, key):
        """
        Set a metadata item.  The body must look like:

        ``{"meta": {<key>: value}}``

        although

        ``{"meta": {<key>: value}, "other": "garbage", "keys": "included"}``

        is ok too.

        All the response messages and codes have been verified as of
        2015-04-23 against Rackspace Nova.
        """
        try:
            content = json.loads(request.content.read())
        except ValueError:
            request.setResponseCode(400)
            return json.dumps(bad_request("Malformed request body", request))

        # more than one key is ok, non-"meta" keys are just ignored
        if 'meta' not in content or not isinstance(content['meta'], dict):
            return json.dumps(bad_request("Malformed request body", request))

        if len(content['meta']) > 1:
            return json.dumps(bad_request(
                "Request body contains too many items", request))

        if key not in content['meta']:
            return json.dumps(bad_request(
                "Request body and URI mismatch", request))

        try:
            self._server.set_metadata_item(key, content['meta'][key])
        except BadRequestError as e:
            return json.dumps(bad_request(e.nova_message, request))
        except LimitError as e:
            return json.dumps(forbidden(e.nova_message, request))

        # no matter how many keys were passed in, only the meta key is
        # returned
        return json.dumps({'meta': content['meta']})



"""
Contains the base Klein app for Mimic.
"""

from klein import Klein


class MimicApp(Klein):
    """
    Base app that extends Klein to override route.
    """
    def route(self, *args, **kwargs):
        """
        Default strict_slashes to False
        """
        kwargs['strict_slashes'] = False
        return super(MimicApp, self).route(*args, **kwargs)
